import torch
import open_clip
from torchvision import transforms
from ForensicHub.tasks.aigc.models.co_spy.cospy_utils import data_augment


# Semantic Detector (Extract semantic features using CLIP)
class SemanticDetector(torch.nn.Module):
    def __init__(self, clip_model="ViT-L-14", pretrained="datacomp_xl_s13b_b90k", dim_clip=768, num_classes=1):
        super(SemanticDetector, self).__init__()
        print(open_clip.list_pretrained())
        # Get the pre-trained CLIP
        model_name = clip_model
        version = pretrained
        self.clip, _, _ = open_clip.create_model_and_transforms(model_name, pretrained=version)
        # # Freeze the CLIP visual encoder
        # self.clip.requires_grad_(False)

        # Classifier
        self.fc = torch.nn.Linear(dim_clip, num_classes)

        # Normalization
        self.mean = [0.5, 0.5, 0.5]
        self.std = [0.5, 0.5, 0.5]

        # Resolution
        self.loadSize = 224
        self.cropSize = 224

        # Data augmentation
        self.blur_prob = 0.5
        self.blur_sig = [0.0, 3.0]
        self.jpg_prob = 0.5
        self.jpg_method = ['cv2', 'pil']
        self.jpg_qual = list(range(30, 101))

        # Define the augmentation configuration
        self.aug_config = {
            "blur_prob": self.blur_prob,
            "blur_sig": self.blur_sig,
            "jpg_prob": self.jpg_prob,
            "jpg_method": self.jpg_method,
            "jpg_qual": self.jpg_qual,
        }

        # Pre-processing
        crop_func = transforms.RandomCrop(self.cropSize)
        flip_func = transforms.RandomHorizontalFlip()
        rz_func = transforms.Resize(self.loadSize)
        aug_func = transforms.Lambda(lambda x: data_augment(x, self.aug_config))

        self.train_transform = transforms.Compose([
            rz_func,
            aug_func,
            crop_func,
            flip_func,
            transforms.ToTensor(),
            transforms.Normalize(mean=self.mean, std=self.std),
        ])

        self.test_transform = transforms.Compose([
            rz_func,
            crop_func,
            transforms.ToTensor(),
            transforms.Normalize(mean=self.mean, std=self.std),
        ])

    def forward(self, x, return_feat=False):
        feat = self.clip.encode_image(x)
        out = self.fc(feat)
        if return_feat:
            return feat, out
        return out

    def save_weights(self, weights_path):
        save_params = {"fc.weight": self.fc.weight.cpu(), "fc.bias": self.fc.bias.cpu()}
        torch.save(save_params, weights_path)

    def load_weights(self, weights_path):
        weights = torch.load(weights_path)
        self.fc.weight.data = weights["fc.weight"]
        self.fc.bias.data = weights["fc.bias"]



import random
from ForensicHub.core.base_dataset import BaseDataset
from ForensicHub.registry import register_dataset, build_from_registry, DATASETS


@register_dataset("CrossDataset")
class CrossDataset(BaseDataset):
    def __init__(self, dataset_config=[], **kwargs):
        super().__init__(path='', **kwargs)
        self.datasets = []
        self.pic_nums = []
        self.dataset_names = []
        self.return_mask = True
        for config in dataset_config:
            config['init_config']['common_transform'] = self.common_transform
            config['init_config']['post_transform'] = self.post_transform
            config['init_config']['post_funcs'] = self.post_funcs
            dataset = build_from_registry(DATASETS, config)
            self.dataset_names.append(config['name'])
            self.datasets.append(dataset)
            self.pic_nums.append(config['pic_nums'])

        # 数据集的最小keys集合
        self.common_keys = self.get_common_keys(self.datasets)
        print(self.common_keys)

    def __len__(self):
        total_samples = sum(self.pic_nums)  # 每个数据集的 pic_num 加起来
        return total_samples

    def _init_dataset_path(self) -> None:
        pass

    def get_common_keys(self, datasets):
        """
        获取所有数据集样本中 key 的最小公共集合
        """
        try:
            # 初始化为第一个数据集的第一个样本的 key 集合
            common_keys = set(datasets[0][0].keys())
            for dataset in datasets[1:]:
                sample = dataset[0]
                common_keys &= set(sample.keys())
            return common_keys
        except Exception as e:
            return set()  # 如果出错，返回空集合

    def __getitem__(self, index):
        # 根据 index 确定应该从哪个数据集抽取图像
        # 累积的 pic_num 可以帮助我们确定当前从哪个数据集抽取
        cumulative_samples = 0

        for i, pic_num in enumerate(self.pic_nums):
            cumulative_samples += pic_num
            if index < cumulative_samples:  # 如果 index 在这个数据集的范围内
                selected_dataset = self.datasets[i]
                # 在当前数据集中随机选择一个样本
                selected_item = random.randint(0, len(selected_dataset) - 1)
                origin_out_dict = selected_dataset[selected_item]
                origin_out_dict = {key: origin_out_dict[key] for key in self.common_keys if key in origin_out_dict}
                origin_out_dict['label'] = origin_out_dict['label'].long()
                return origin_out_dict

        raise IndexError("Index out of range")

    def __str__(self):
        info = f"<=== CrossDataset with {len(self.datasets)} datasets: {self.dataset_names} ===>\n"
        for i, ds in enumerate(self.datasets):
            info += f"\n[Dataset {i} - {self.dataset_names[i]}]\n"
            info += str(ds) + "\n"
        info += f"\nTotal samples per epoch: {self.__len__():,}\n"
        info += f"<================================================>\n"
        return info



# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------
from os.path import isfile, join
from PIL import Image
from torchvision import transforms
import numpy as np
import abc
import cv2
import torch.utils.data as data
import torch.nn.functional as F
import random
random.seed(1234567890)
from random import randrange
import torch.nn as nn
import torch
import imageio
import time
import math
import torch

class BaseData(data.Dataset):
    '''
        The dataset used for the IFDL dataset.
    '''
    def __init__(self, args):
        super(BaseData, self).__init__()
        self.crop_size = args.crop_size
        self.file_path = '/user/guoxia11/cvlshare/cvl-guoxia11/IMDL/REAL'
        self.file_path_fake = '/user/guoxia11/cvlshare/cvl-guoxia11/IMDL/FAKE'     

        # Real and Fake images.
        self.image_names = []
        self.image_class = self._img_list_retrieve()
        for idx, _ in enumerate(self.image_class):
            self.image_names += _

    def __getitem__(self, index):
        res = self.get_item(index)
        return res

    def __len__(self):
        return len(self.image_names)

    @abc.abstractmethod
    def _img_list_retrieve():
        pass

    def _resize_func(self, input_img):
        '''resize the input image into the crop size.'''
        input_img = Image.fromarray(input_img)
        input_img = input_img.resize(self.crop_size, resample=Image.BICUBIC)
        input_img = np.asarray(input_img)
        return input_img

    def get_image(self, image_name, aug_index=None):
        '''transform the image.'''
        image = imageio.imread(image_name)
        if image.shape[-1] == 4:
            image = self.rgba2rgb(image)
        image = self._resize_func(image)
        image = image.astype(np.float32) / 255.
        image = torch.from_numpy(image)
        return image.permute(2, 0, 1)

    def rgba2rgb(self, rgba, background=(255, 255, 255)):
        '''
            turn rgba to rgb.
        '''
        row, col, ch = rgba.shape
        rgb = np.zeros((row, col, 3), dtype='float32')
        r, g, b, a = rgba[:, :, 0], rgba[:, :, 1], rgba[:, :, 2], rgba[:, :, 3]
        a = np.asarray(a, dtype='float32') / 255.0

        R, G, B = background
        rgb[:, :, 0] = r * a + (1.0 - a) * R
        rgb[:, :, 1] = g * a + (1.0 - a) * G
        rgb[:, :, 2] = b * a + (1.0 - a) * B

        return np.asarray(rgb, dtype='uint8')

    def generate_4masks(self, mask):
        '''generate 4 masks at different scale.'''
        crop_height, crop_width = self.crop_size
        ma_height, ma_width = mask.shape[:2]
        mask_pil = Image.fromarray(mask)

        if ma_height != crop_height or ma_width != crop_width:
            mask_pil = mask_pil.resize(self.crop_size, resample=Image.BICUBIC)

        (width2, height2) = (mask_pil.width // 2, mask_pil.height // 2)
        (width3, height3) = (mask_pil.width // 4, mask_pil.height // 4)
        (width4, height4) = (mask_pil.width // 8, mask_pil.height // 8)

        mask2 = mask_pil.resize((width2, height2))
        mask3 = mask_pil.resize((width3, height3))
        mask4 = mask_pil.resize((width4, height4))

        mask = np.asarray(mask_pil)
        mask = mask.astype(np.float32) / 255
        mask[mask > 0.5] = 1
        mask[mask <= 0.5] = 0

        mask2 = np.asarray(mask2).astype(np.float32) / 255
        mask2[mask2 > 0.5] = 1
        mask2[mask2 <= 0.5] = 0

        mask3 = np.asarray(mask3).astype(np.float32) / 255
        mask3[mask3 > 0.5] = 1
        mask3[mask3 <= 0.5] = 0

        mask4 = np.asarray(mask4).astype(np.float32) / 255
        mask4[mask4 > 0.5] = 1
        mask4[mask4 <= 0.5] = 0

        mask = torch.from_numpy(mask)
        mask2 = torch.from_numpy(mask2)
        mask3 = torch.from_numpy(mask3)
        mask4 = torch.from_numpy(mask4)

        # print(mask.size(), mask2.size(), mask3.size(), mask4.size())

        return mask, mask2, mask3, mask4

    def get_mask(self, image_name, cls, aug_index=None):
        '''given the cls, we return the mask.'''
        # authentic
        if cls in [0,1,2,3,4]:
            mask = self.load_mask('', real=True, aug_index=aug_index)
            return_res = [0,0,0,0]
        
        # splice
        elif cls == 5:
            if '.jpg' in image_name:
                mask_name = image_name.replace('fake', 'mask').replace('.jpg', '.png')
            else:
                mask_name = image_name.replace('fake', 'mask').replace('.tif', '.png')
            mask = self.load_mask(mask_name, aug_index=aug_index)
            return_res = [1,1,1,cls - 4]
        
        # inpainting
        elif cls == 6:
            mask_name = image_name.replace('/fake/', '/mask/').replace('.jpg', '.png')
            mask = self.load_mask(mask_name, aug_index=aug_index)
            return_res = [1,1,1,cls - 4]

        # copy-move
        elif cls == 7:
            mask_name = image_name.replace('.png', '_mask.png')
            mask_name = mask_name.replace('CopyMove', 'CopyMove_mask')
            mask = self.load_mask(mask_name, aug_index=aug_index)
            return_res = [1,1,1,cls - 4]

        # faceshifter
        elif cls == 8:  
            image_id  = image_name.split('/')[-1].split('.')[0]
            mask_name = image_name.replace(image_id, f'mask/{image_id}_mask')
            mask = self.load_mask(mask_name, aug_index=aug_index)
            return_res = [1,2,2,cls - 4]

        # STGAN
        elif cls == 9: 
            image_id  = image_name.split('/')[-1].split('.')[0]
            mask_name = image_name.replace('fake', 'mask').replace(image_id, f'{image_id}_label')
            mask = self.load_mask(mask_name, aug_index=aug_index)
            return_res = [1,2,2,cls - 4]

        ## they are star2, hisd, stylegan2, stylegan3, ddpm, ddim, latent, guided
        elif cls in [10,11,12,13,14,15,16,17]:  
            mask = self.load_mask('', real=False, full_syn=True, aug_index=aug_index)
            if cls in [10,11]:
                return_res = [2,3,3,cls-4]
            elif cls in [12,13]:
                return_res = [2,3,4,cls-4]
            elif cls in [14,15]:
                return_res = [2,4,5,cls-4]
            elif cls in [16,17]:
                return_res = [2,4,6,cls-4]
        else:
            print(cls, index)
            raise Exception('class is not defined!')

        return mask, return_res

    def load_mask(self, mask_name, real=False, full_syn=False, gray=True, aug_index=None):
        '''binarize the mask, given the mask_name.'''
        if real:
            mask = np.zeros(self.crop_size)
        else:
            if not full_syn:
                mask = imageio.imread(mask_name) if not gray else np.asarray(Image.open(mask_name).convert('RGB').convert('L'))
                mask = mask.astype(np.float32)
            else:
                mask = np.ones(self.crop_size)

        mask = self.generate_4masks(mask)
        return mask

    def get_cls(self, image_name):
        '''return the forgery/authentic cls given the image_name.'''
        if '/authentic/' in image_name:
            return_cls = 0
        elif '/REAL/LSUN/' in image_name:
            return_cls = 0
        elif '/afhq_v2/' in image_name:
            return_cls = 1
        elif '/CelebAHQ/' in image_name:
            return_cls = 2
        elif '/FFHQ/' in image_name:
            return_cls = 3
        elif '/Youtube' in image_name:
            return_cls = 4
        elif '/splice' in image_name:
            return_cls = 5
        elif '/Inpainting' in image_name:
            return_cls = 6
        elif '/CopyMove' in image_name:
            return_cls = 7
        elif '/FaShifter' in image_name:
            return_cls = 8
        elif '/STGAN' in image_name:
            return_cls = 9
        elif '/Star2' in image_name:
            return_cls = 10
        elif '/HiSD' in image_name:
            return_cls = 11
        elif '/STYL2' in image_name:
            return_cls = 12
        elif '/STYL3' in image_name:
            return_cls = 13
        elif '/DDPM_' in image_name:
            return_cls = 14
        elif '/DDIM_' in image_name:
            return_cls = 15
        elif '/D_latent' in image_name:
            return_cls = 16
        elif '/GLIDE/' in image_name:
            return_cls = 17
        else:
            print(image_name)
            raise ValueError 
        return return_cls

class TrainData(BaseData):
    '''
        The dataset used for the IFDL dataset.
    '''
    def __init__(self, args):
        self.is_train = True
        self.val_num  = 90000
        super(TrainData, self).__init__(args)

    def img_retrieve(self, file_text, file_folder, real=True):
        '''
            Parameters:
                file_text: str, text file for images.
                file_folder: str, images folder.
            Returns:
                the image list.
        '''
        result_list = []
        val_num   = self.val_num * 3 if file_text in ["Youtube", "FaShifter"] else self.val_num
        data_path = self.file_path if real else self.file_path_fake

        data_text = join(data_path, file_text)
        data_path = join(data_path, file_folder)

        file_handler = open(data_text)
        contents = file_handler.readlines()
        if self.is_train:
            contents_lst = contents[:val_num]
        else:
            contents_lst = contents[val_num:]

        for content in contents_lst:
            if '.npy' not in content and 'mask' not in content:
                img_name = content.strip()
                img_name = join(data_path, img_name)
                result_list.append(img_name)
        file_handler.close()

        ## only truncate the val_num images. 
        if len(result_list) < val_num:
            mul_factor  = (val_num//len(result_list)) + 2
            result_list = result_list * mul_factor
        result_list = result_list[-val_num:]
        return result_list

    def get_item(self, index):
        '''
            given the index, this function returns the image with the forgery mask
            this function calls get_image, get_mask for the image and mask torch tensor.
        '''
        image_name = self.image_names[index]
        cls = self.get_cls(image_name)

        # image and mask
        aug_index = randrange(0, 8)
        image = self.get_image(image_name, aug_index)
        mask, return_res = self.get_mask(image_name, cls, aug_index)

        return image, mask, return_res[0], return_res[1], return_res[2], return_res[3]

    def _img_list_retrieve(self):
        '''Returns image list for different authentic and forgery image.'''
        authentic_names = self.img_retrieve('authentic.txt', 'authentic')
        splice_names     = self.img_retrieve('splice_randmask.txt', 'splice_randmask/fake',False)
        inpainting_names = self.img_retrieve('Inpainting.txt', 'Inpainting/fake', False)
        copymove_names   = self.img_retrieve('copy_move.txt', 'CopyMove', False)
        STGAN_names   = self.img_retrieve('STGAN.txt', 'STGAN/fake', False)
        FaShifter_names  = self.img_retrieve('FaShifter.txt', 'FaShifter', False)
        return [authentic_names, splice_names, inpainting_names, copymove_names, STGAN_names, FaShifter_names]

class ValData(BaseData):
    '''
        The dataset used for the IFDL dataset.
    '''
    def __init__(self, args):
        self.is_train  = False
        self.val_num   = 900
        super(ValData, self).__init__(args)

    def img_retrieve(self, file_text, file_folder, real=True):
        '''
            Parameters:
                file_text: str, text file for images.
                file_folder: str, images folder.
            Returns:
                the image list.
        '''
        result_list = []
        val_num   = self.val_num * 3 if file_text in ["Youtube", "FaShifter"] else self.val_num
        data_path = self.file_path if real else self.file_path_fake

        data_text = join(data_path, file_text)
        data_path = join(data_path, file_folder)

        file_handler = open(data_text)
        contents = file_handler.readlines()
        for content in contents[-val_num:]:
            if '.npy' not in content and 'mask' not in content:
                img_name = content.strip()
                img_name = join(data_path, img_name)
                result_list.append(img_name)
        file_handler.close()

        ## only truncate the val_num images. 
        if len(result_list) < val_num:
            mul_factor  = (val_num//len(result_list)) + 2
            result_list = result_list * mul_factor
        result_list = result_list[-val_num:]
        return result_list
    
    def get_item(self, index):
        '''
            given the index, this function returns the image with the forgery mask
            this function calls get_image, get_mask for the image and mask torch tensor.
        '''
        image_name = self.image_names[index]
        cls = self.get_cls(image_name)

        # image
        image = self.get_image(image_name)
        mask, return_res = self.get_mask(image_name, cls)

        return image, mask, return_res[0], return_res[1], return_res[2], return_res[3], image_name

    def _img_list_retrieve(self):
        '''Returns image list for different authentic and forgery image.'''
        STGAN_names   = self.img_retrieve('STGAN.txt', 'STGAN/fake', False)
        FaShifter_names  = self.img_retrieve('FaShifter.txt', 'FaShifter', False)
        return [STGAN_names, FaShifter_names]



import torch
import torch.nn as nn
import torch.nn.functional as F

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model
from IMDLBenCo.modules.extractors.sobel import SobelFilter
from IMDLBenCo.modules.extractors.bayar_conv import BayerConv
from IMDLBenCo.modules.extractors.high_frequency_feature_extraction import (
    FFTExtractor, DCTExtractor
)


class DoubleConv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class UNetBackbone(nn.Module):
    def __init__(self, in_channels=3, base_channels=64):
        super().__init__()

        self.inc = DoubleConv(in_channels, base_channels)
        self.down1 = nn.Sequential(nn.MaxPool2d(2), DoubleConv(base_channels, base_channels * 2))
        self.down2 = nn.Sequential(nn.MaxPool2d(2), DoubleConv(base_channels * 2, base_channels * 4))
        self.down3 = nn.Sequential(nn.MaxPool2d(2), DoubleConv(base_channels * 4, base_channels * 8))
        self.down4 = nn.Sequential(nn.MaxPool2d(2), DoubleConv(base_channels * 8, base_channels * 8))

        self.up1 = nn.ConvTranspose2d(base_channels * 8, base_channels * 8, kernel_size=2, stride=2)
        self.conv1 = DoubleConv(base_channels * 16, base_channels * 4)
        self.up2 = nn.ConvTranspose2d(base_channels * 4, base_channels * 4, kernel_size=2, stride=2)
        self.conv2 = DoubleConv(base_channels * 8, base_channels * 2)
        self.up3 = nn.ConvTranspose2d(base_channels * 2, base_channels * 2, kernel_size=2, stride=2)
        self.conv3 = DoubleConv(base_channels * 4, base_channels)
        self.up4 = nn.ConvTranspose2d(base_channels, base_channels, kernel_size=2, stride=2)
        self.conv4 = DoubleConv(base_channels * 2, base_channels)

        self.out_channels = base_channels

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)

        x = self.up1(x5)
        x = self.conv1(torch.cat([x, x4], dim=1))
        x = self.up2(x)
        x = self.conv2(torch.cat([x, x3], dim=1))
        x = self.up3(x)
        x = self.conv3(torch.cat([x, x2], dim=1))
        x = self.up4(x)
        x = self.conv4(torch.cat([x, x1], dim=1))

        return x


class UNet(BaseModel):
    def __init__(self,
                 input_head=None,
                 output_type='mask',
                 pretrained=False,
                 image_size=256,
                 num_channels=3):
        super(UNet, self).__init__()

        self.output_type = output_type
        self.input_head = input_head

        self.model = UNetBackbone(in_channels=num_channels + 3 if input_head else 3)
        out_channels = self.model.out_channels

        if output_type == 'label':
            self.head = nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
                nn.Linear(out_channels, 1)
            )
        elif output_type == 'mask':
            self.head = nn.Sequential(
                nn.Conv2d(out_channels, 1, kernel_size=1),
                nn.Upsample(size=(image_size, image_size), mode='bilinear', align_corners=False)
            )
        else:
            raise ValueError(f"Unsupported output_type: {output_type}")

    def forward(self, image, *args, **kwargs):
        if self.input_head is not None:
            feature = self.input_head(image)
            x = torch.cat([image, feature], dim=1)
        else:
            x = image

        features = self.model(x)
        out = self.head(features)

        if self.output_type == 'label':
            if len(out.shape) == 2:
                out = out.squeeze(dim=1)
            loss = F.binary_cross_entropy_with_logits(out, kwargs['label'].float())
            pred = out.sigmoid()
        else:
            loss = F.binary_cross_entropy_with_logits(out, kwargs['mask'].float())
            pred = out.sigmoid()

        return {
            "backward_loss": loss,
            f"pred_{self.output_type}": pred,
            "visual_loss": {"combined_loss": loss}
        }


@register_model("UNetBase")
class UNetBase(UNet):
    def __init__(self, output_type='mask', image_size=256):
        super().__init__(input_head=None, output_type=output_type,
                         image_size=image_size, num_channels=0)


@register_model("SobelUNet")
class SobelUNet(UNet):
    def __init__(self, output_type='mask', image_size=256):
        super().__init__(input_head=SobelFilter(), output_type=output_type,
                         image_size=image_size, num_channels=1)


@register_model("BayerUNet")
class BayerUNet(UNet):
    def __init__(self, output_type='mask', image_size=256):
        super().__init__(input_head=BayerConv(), output_type=output_type,
                         image_size=image_size, num_channels=3)


@register_model("FFTUNet")
class FFTUNet(UNet):
    def __init__(self, output_type='mask', image_size=256):
        super().__init__(input_head=FFTExtractor(), output_type=output_type,
                         image_size=image_size, num_channels=3)


@register_model("DCTUNet")
class DCTUNet(UNet):
    def __init__(self, output_type='mask', image_size=256):
        super().__init__(input_head=DCTExtractor(), output_type=output_type,
                         image_size=image_size, num_channels=3)



import torch
import torch.nn as nn
from torch.nn import functional as F


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = conv1x1(planes, planes * self.expansion)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class VAEReconEncoder(nn.Module):
    def __init__(self, vae, block=Bottleneck):
        super(VAEReconEncoder, self).__init__()

        # Define the ResNet model
        self.inplanes = 64
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1, bias=False)
        # self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        # ResNet-50 is [3, 4, 6, 3]
        self.layer1 = self._make_layer(block, 64 , 3)
        self.layer2 = self._make_layer(block, 128, 4, stride=2)
        # self.layer3 = self._make_layer(block, 256, 6, stride=2)
        # self.layer4 = self._make_layer(block, 512, 3, stride=2)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        # Kaiming initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        
        # Load the VAE model
        self.vae = vae

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def reconstruct(self, x):
        with torch.no_grad():
            # `.sample()` means to sample a latent vector from the distribution
            # `.mean` means to use the mean of the distribution
            latent = self.vae.encode(x).latent_dist.mean
            decoded = self.vae.decode(latent).sample
        return decoded

    def forward(self, x):
        # Reconstruct
        x_recon = self.reconstruct(x)
        # Compute the artifacts
        x = x - x_recon

        # Scale the artifacts
        x = x / 7. * 100.

        # Forward pass
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)

        return x



import timm
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import segmentation_models_pytorch as smp

from ForensicHub.registry import register_model
from ForensicHub.core.base_model import BaseModel

'''
Cross-Attention Based Two-Branch Networks for Document Image Forgery Localization in the Metaverse
https://dl.acm.org/doi/abs/10.1145/3686158
'''


class CSA(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.cSE = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, in_channels, 1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return x * self.cSE(x)


class RSA(nn.Module):
    def __init__(self, ):
        super().__init__()
        self.conv = nn.Conv2d(1, 1, 7, 1, 3)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        return x * self.sigmoid(self.conv(x.mean(1, keepdim=True) + x.max(1, keepdim=True).values))


class CAFM(nn.Module):
    def __init__(self, in_channels, resize=1):
        super().__init__()
        self.csa = CSA(in_channels)
        self.rsa = RSA()
        self.conv = nn.Sequential(nn.Conv2d(in_channels, in_channels, 1, 1, 0), nn.BatchNorm2d(in_channels), nn.ReLU())
        self.resize = resize

    def forward(self, xc, xt):
        if self.resize == 1:
            return self.conv(self.csa(xc) + self.rsa(xt))
        else:
            return F.interpolate(self.conv(self.csa(xc) + self.rsa(xt)), scale_factor=self.resize, mode='bilinear')


class SoftDiceLossV1(nn.Module):
    '''
    soft-dice loss, useful in binary segmentation
    '''

    def __init__(self,
                 p=1,
                 smooth=1):
        super(SoftDiceLossV1, self).__init__()
        self.p = p
        self.smooth = smooth

    def forward(self, logits, labels):
        '''
        inputs:
            logits: tensor of shape (N, H, W, ...)
            label: tensor of shape(N, H, W, ...)
        output:
            loss: tensor of shape(1, )
        '''
        probs = F.softmax(logits, 1)[:, 1]
        numer = (probs * labels).sum()
        denor = (probs.pow(self.p) + labels.pow(self.p)).sum()
        loss = 1. - (2 * numer + self.smooth) / (denor + self.smooth)
        return loss


@register_model("CAFTB_Net")
class CAFTB_Net(BaseModel):
    def __init__(self, cnn_model_name='resnetv2_50d_gn.ah_in1k',
                 vit_pretrained_path='/mnt/data1/public_datasets/Doc/Hub/caftb-net/segformer-b5-640x640-ade-160k'):
        super(CAFTB_Net, self).__init__()
        self.cnn = timm.create_model(cnn_model_name, pretrained=False, features_only=True)
        self.cnn.stem_conv3 = nn.Conv2d(32, 64, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)
        self.cnn.stages_3.blocks[0].downsample.pool = nn.Identity()
        self.cnn.stages_3.blocks[0].conv2 = nn.Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1),
                                                      bias=False)
        self.vit = smp.from_pretrained(vit_pretrained_path).encoder.eval()
        channels = [64, 256, 512, 1024]
        self.convert0 = nn.Sequential(nn.Conv2d(2048, channels[3], 1, 1, 0), nn.ReLU())
        self.convert1 = nn.Sequential(nn.Conv2d(128, channels[1], 1, 1, 0), nn.ReLU())
        self.convert2 = nn.Sequential(nn.Conv2d(320, channels[2], 1, 1, 0), nn.ReLU())
        self.convert3 = nn.Sequential(nn.Conv2d(512, channels[3], 1, 1, 0), nn.ReLU())
        self.cafms = nn.ModuleList([CAFM(channels[i], 2 ** i) for i in range(4)])
        self.decoder = nn.Sequential(nn.Conv2d(sum(channels), 64, 3, 1, 1), nn.ReLU(), nn.Conv2d(64, 2, 1, 1, 0))
        self.dice_loss = SoftDiceLossV1()

    def vit_forward(self, x):
        rsts = []
        x = self.vit.patch_embed1(x)
        x = self.vit.block1(x)
        x = self.vit.norm1(x)
        x = x.contiguous()
        rsts.append(x)

        x = self.vit.patch_embed2(x)
        x = self.vit.block2(x)
        x = self.vit.norm2(x)
        x = x.contiguous()
        rsts.append(self.convert1(x))

        x = self.vit.patch_embed3(x)
        x = self.vit.block3(x)
        x = self.vit.norm3(x)
        x = x.contiguous()
        rsts.append(self.convert2(x))

        x = self.vit.patch_embed4(x)
        x = self.vit.block4(x)
        x = self.vit.norm4(x)
        x = x.contiguous()
        rsts.append(self.convert3(x))

        return rsts

    def cal_seg_loss(self, pred, gt):
        h, w = gt.shape[-2:]
        pred = F.interpolate(pred, size=(h, w), mode='bilinear')
        ce_loss = (F.cross_entropy(pred, gt) * 0.3 + self.dice_loss(pred, gt) * 0.7)
        return ce_loss, pred

    def forward(self, image, mask, **kwargs):
        x = image
        mask = mask.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        xc = self.cnn(x)
        xc = [xc[0], xc[1], xc[2], self.convert0(xc[4])]
        xt = self.vit_forward(x)
        xc = torch.cat([self.cafms[i](xc[i], xt[i]) for i in range(4)], 1)
        output = self.decoder(xc)
        seg_loss, output = self.cal_seg_loss(output, mask)
        pred_mask = F.softmax(output, dim=1)[:,1:]
        output_dict = {
            "backward_loss": seg_loss,
            "pred_mask": pred_mask,
            "visual_loss": {
                "seg_loss": seg_loss,
                "combined_loss": seg_loss
            },
            "visual_image": {
                "pred_mask": pred_mask,
            }
        }
        return output_dict


if __name__ == "__main__":
    img = torch.ones((1, 3, 512, 512))
    mask = torch.ones((1, 1, 512, 512), dtype=torch.int64)
    model = CAFTB_Net()
    pred = model(img, mask)
    print(pred)



import torch
import torch.nn as nn

from .clip import clip 
from .clip.simple_tokenizer import SimpleTokenizer as _Tokenizer


_tokenizer = _Tokenizer()


class TextEncoder(nn.Module):
    def __init__(self, clip_model):
        super().__init__()
        self.transformer = clip_model.transformer
        self.positional_embedding = clip_model.positional_embedding
        self.ln_final = clip_model.ln_final
        self.text_projection = clip_model.text_projection
        self.dtype = clip_model.dtype

    def forward(self, prompts, tokenized_prompts):
        x = prompts + self.positional_embedding.type(self.dtype)
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)[1]
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x).type(self.dtype)

        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), tokenized_prompts.argmax(dim=-1)] @ self.text_projection

        return x


class LanguageGuidedAlignment(nn.Module):
    def __init__(self, clip_model, classnames=["real", "synthetic"], args=None):
        super().__init__()
        n_cls = len(classnames)
        n_ctx = args.num_context_embedding
        ctx_init = args.init_context_embedding
        
        # n_ctx = 4
        # ctx_init = ""
        
        dtype = clip_model.dtype
        ctx_dim = clip_model.ln_final.weight.shape[0]

        if ctx_init:
            # use given words to initialize context vectors
            ctx_init = ctx_init.replace("_", " ")
            n_ctx = len(ctx_init.split(" "))
            prompt = clip.tokenize(ctx_init)
            with torch.no_grad():
                embedding = clip_model.token_embedding(prompt).type(dtype)
            ctx_vectors = embedding[0, 1 : 1 + n_ctx, :]
            prompt_prefix = ctx_init
        else:
            # random initialization
            ctx_vectors = torch.empty(n_ctx, ctx_dim, dtype=dtype)
            nn.init.normal_(ctx_vectors, std=0.02)
            prompt_prefix = " ".join(["X"] * n_ctx)

        print(f'Initial context: "{prompt_prefix}"')
        print(f"Number of context words (tokens): {n_ctx}")
        # print("clip_model.visual:", clip_model.visual)
        # print("clip_model.visual.output_dim:", clip_model.visual.output_dim)


        self.ctx = nn.Parameter(ctx_vectors)

        classnames = [name.replace("_", " ") for name in classnames]
        name_lens = [len(_tokenizer.encode(name)) for name in classnames]
        prompts = [prompt_prefix + " " + name + "." for name in classnames]

        tokenized_prompts = torch.cat([clip.tokenize(p) for p in prompts])  # (n_cls, n_tkn)
        with torch.no_grad():
            embedding = clip_model.token_embedding(tokenized_prompts).type(dtype)

        # These token vectors will be saved when in save_model(),
        # but they should be ignored in load_model() as we want to use
        # those computed using the current class names
        self.register_buffer("token_prefix", embedding[:, :1, :])  # SOS
        self.register_buffer("token_suffix", embedding[:, 1 + n_ctx :, :])  # CLS, EOS

        self.n_cls = n_cls
        self.n_ctx = n_ctx
        self.tokenized_prompts = tokenized_prompts  # torch.Tensor
        self.name_lens = name_lens
        
        # patch-basaed enhancer in LGA
        d_model = clip_model.ln_final.weight.shape[0]
        d_ffn = d_model * 4
        self.patch_basaed_enhancer = nn.MultiheadAttention(d_model, num_heads=args.num_heads)
        self.norm1 = nn.LayerNorm(d_model)
        # FFN in patch-basaed enhancer
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = nn.ReLU()
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm2 = nn.LayerNorm(d_model)

    def forward_FFN(self, tgt):
        tgt2 = self.linear2(self.activation(self.linear1(tgt)))
        tgt = tgt + tgt2
        tgt = self.norm2(tgt)
        return tgt
    
    def construct_prompts(self, ctx, prefix, suffix, label=None):
        # dim0 is either batch_size (during training) or n_cls (during testing)
        # ctx: context tokens, with shape of (dim0, n_ctx, ctx_dim)
        # prefix: the sos token, with shape of (n_cls, 1, ctx_dim)
        # suffix: remaining tokens, with shape of (n_cls, *, ctx_dim)

        if label is not None:
            prefix = prefix[label]
            suffix = suffix[label]

        prompts = torch.cat(
            [
                prefix,  # (dim0, 1, dim)
                ctx,     # (dim0, n_ctx, dim)
                suffix,  # (dim0, *, dim)
            ],
            dim=1,
        )

        return prompts

    def forward(self, im_features):
        prefix = self.token_prefix
        suffix = self.token_suffix
        ctx = self.ctx                     # (n_ctx, ctx_dim)

        # patch-basaed enhancer
        tgt = ctx[:, None].repeat_interleave(im_features.shape[0], dim=1)
        tgt2 = self.patch_basaed_enhancer(tgt, im_features.transpose(0, 1), im_features.transpose(0, 1))[0]
        tgt = tgt + tgt2
        tgt = self.norm1(tgt)
        tgt = self.forward_FFN(tgt)

        ctx_shifted = tgt.transpose(0, 1)
        
        # Use instance-conditioned context tokens for all classes
        prompts = []
        for ctx_shifted_i in ctx_shifted:
            ctx_i = ctx_shifted_i.unsqueeze(0).expand(self.n_cls, -1, -1)
            pts_i = self.construct_prompts(ctx_i, prefix, suffix)  # (n_cls, n_tkn, ctx_dim)
            prompts.append(pts_i)
        prompts = torch.stack(prompts)
        
        return prompts


class CLIPModel(nn.Module):
    def __init__(self, name, args=None):
        super(CLIPModel, self).__init__()
        
        # init backbone with forgery-aware adapter
        self.clip_model = clip.load(name, device="cpu", args=args)[0] # self.preprecess will not be used during training, which is handled in Dataset class 
        # init language guided alignment
        self.language_guided_alignment = LanguageGuidedAlignment(self.clip_model, classnames=["real", "fake"], args=args)
        
        self.tokenized_prompts = self.language_guided_alignment.tokenized_prompts
        self.image_encoder = self.clip_model.visual
        self.text_encoder = TextEncoder(self.clip_model)
        self.logit_scale = self.clip_model.logit_scale
        self.dtype = self.clip_model.dtype
        self.num_classes = args.num_classes

        # text-guided interactor in LGA
        d_model = self.clip_model.ln_final.weight.shape[0]
        d_ffn = d_model * 4
        self.text_guided_interactor = nn.MultiheadAttention(d_model, num_heads=args.num_heads)
        self.norm1 = nn.LayerNorm(d_model)
        # FFN in text-guided interactor
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = nn.ReLU()
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm2 = nn.LayerNorm(d_model)

    def forward_FFN(self, tgt):
        tgt2 = self.linear2(self.activation(self.linear1(tgt)))
        tgt = tgt + tgt2
        tgt = self.norm2(tgt)
        return tgt

    def forward(self, image):
        tokenized_prompts = self.tokenized_prompts
        logit_scale = self.logit_scale.exp()

        image_features = self.image_encoder(image.type(self.dtype), return_full=True)
        image_features_nrom = image_features / image_features.norm(dim=-1, keepdim=True)

        prompts = self.language_guided_alignment(image_features)
        
        # Eq.(1)
        logits = []
        text_feature_list = []
        for pts_i, imf_i in zip(prompts, image_features_nrom):
            text_features = self.text_encoder(pts_i, tokenized_prompts)
            text_feature_list.append(text_features)
            text_features_norm = text_features / text_features.norm(dim=-1, keepdim=True)
            l_i = logit_scale * imf_i[0] @ text_features_norm.t()
            logits.append(l_i)
        logits = torch.stack(logits)

        # Eq.(9)
        # text-guided interactor
        text_features = torch.stack(text_feature_list, dim=1)
        tgt = image_features[:, 1:].transpose(0, 1)
        tgt2 = self.text_guided_interactor(tgt, text_features, text_features)[0]
        tgt = tgt + tgt2
        tgt = self.norm1(tgt)
        tgt = self.forward_FFN(tgt)
        
        aug_image_features = tgt.transpose(0, 1).mean(dim=1)
        aug_image_features = aug_image_features / aug_image_features.norm(dim=-1, keepdim=True)
        text_features_norm = text_features / text_features.norm(dim=-1, keepdim=True)
        aug_logits = []
        for pts_i, imf_i in zip(aug_image_features, text_features_norm.transpose(0, 1)):
            aug_logits.append(logit_scale * pts_i @ imf_i.t())
        aug_logits = torch.stack(aug_logits)
        
        return logits + aug_logits



import torch
import torch.nn as nn
import torch.nn.functional as F
import timm
from .extractors.fph import FPH

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model
from IMDLBenCo.modules.extractors.high_frequency_feature_extraction import (
    FFTExtractor,
    DCTExtractor
)
from IMDLBenCo.modules.extractors.sobel import SobelFilter
from IMDLBenCo.modules.extractors.bayar_conv import BayerConv


class XceptionNet(BaseModel):
    def __init__(self,
                 input_head=None,
                 output_type='label',
                 pretrained=True,
                 image_size=256,
                 num_channels=3):
        super(XceptionNet, self).__init__()
        self.model = timm.create_model('legacy_xception', pretrained=pretrained)
        self.output_type = output_type
        self.input_head = input_head

        # 替换第一层以适配 input_head 融合
        if input_head is not None:
            original_first_layer = self.model.conv1
            new_first_layer = nn.Conv2d(num_channels + 3, original_first_layer.out_channels,
                                        kernel_size=original_first_layer.kernel_size,
                                        stride=original_first_layer.stride,
                                        padding=original_first_layer.padding, bias=False)
            new_first_layer.weight.data[:, :3, :, :] = original_first_layer.weight.data.clone()[:, :3, :, :]
            if num_channels > 0:
                new_first_layer.weight.data[:, 3:, :, :] = torch.nn.init.kaiming_normal_(
                    new_first_layer.weight[:, 3:, :, :]
                )
            self.model.conv1 = new_first_layer

        # 提取分类头的输出维度
        out_channels = self.model.get_classifier().in_features

        # 重建 head
        if output_type == 'label':
            self.head = nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
                nn.Linear(out_channels, 1)
            )
        elif output_type == 'mask':
            self.head = nn.Sequential(
                nn.Conv2d(out_channels, out_channels // 2, kernel_size=3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(out_channels // 2, out_channels // 4, kernel_size=3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(out_channels // 4, 1, kernel_size=1),
                nn.Upsample(size=(image_size, image_size), mode='bilinear', align_corners=False)
            )
        else:
            raise ValueError(f"Unsupported output_type: {output_type}")

    def forward(self, image, *args, **kwargs):
        if self.input_head is not None:
            feature = self.input_head(image)
            x = torch.cat([image, feature], dim=1)
        else:
            x = image

        # 手动分离出 features，兼容 legacy xception
        features = self.model.forward_features(x)
        out = self.head(features)

        if self.output_type == 'label':
            if len(out.shape) == 2:
                out = out.squeeze(dim=1)
            loss = F.binary_cross_entropy_with_logits(out, kwargs['label'].float())
            pred = out.sigmoid()
        else:
            loss = F.binary_cross_entropy_with_logits(out, kwargs['mask'].float())
            pred = out.sigmoid()
        return {
            "backward_loss": loss,
            f"pred_{self.output_type}": pred,
            "visual_loss": {"combined_loss": loss}
        }


@register_model("Xception")
class Xception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type, pretrained=pretrained,
                         image_size=image_size, num_channels=0)


@register_model("SobelXception")
class SobelXception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=SobelFilter(), output_type=output_type, pretrained=pretrained,
                         image_size=image_size, num_channels=1)


@register_model("BayerXception")
class BayerXception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=BayerConv(), output_type=output_type, pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("FFTXception")
class FFTXception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=FFTExtractor(), output_type=output_type, pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("DCTXception")
class DCTXception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=DCTExtractor(), output_type=output_type, pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("QtXception")
class QtXception(XceptionNet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type,
                         pretrained=pretrained, image_size=image_size, num_channels=0)
        self.fph = FPH()

        # Xception的结构分解
        # stem部分：conv1 + bn1 + act1 + conv2 + bn2 + act2
        self.stem = nn.Sequential(
            self.model.conv1,
            self.model.bn1,
            self.model.act1,
            self.model.conv2,
            self.model.bn2,
            self.model.act2
        )

        # 主体blocks部分：从block1到block12
        self.blocks = nn.Sequential(
            self.model.block1,
            self.model.block2,
            self.model.block3,
            self.model.block4,
            self.model.block5,
            self.model.block6,
            self.model.block7,
            self.model.block8,
            self.model.block9,
            self.model.block10,
            self.model.block11,
            self.model.block12
        )

        self.tail = nn.Sequential(
            self.model.conv3,
            self.model.bn3,
            self.model.act3,
            self.model.conv4,
            self.model.bn4,
            self.model.act4,
        )

        # 通道融合层（Xception的stem输出通道通常是64）
        in_channels = 64 + 256  # stem输出64通道 + FPH的256通道
        self.fusion_conv = nn.Conv2d(in_channels, 64, kernel_size=1)

    def forward(self, image, dct, qt, *args, **kwargs):
        dct = dct.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        # FPH特征提取 (B, 256, H/8, W/8)
        x_aux = self.fph(dct, qt)

        # 主干前部分
        x = self.stem(image)

        # 尺寸对齐 (Xception的stem输出尺寸是输入的1/4)
        if x.shape[-2:] != x_aux.shape[-2:]:
            x_aux = F.interpolate(x_aux, size=x.shape[-2:], mode='bilinear', align_corners=False)

        # 拼接并融合
        x = torch.cat([x, x_aux], dim=1)
        x = self.fusion_conv(x)

        # 继续主干网络
        x = self.blocks(x)
        x = self.tail(x)
        out = self.head(x)

        if self.output_type == 'label':
            if len(out.shape) == 2:
                out = out.squeeze(dim=1)
            loss = F.binary_cross_entropy_with_logits(out, kwargs['label'].float())
            pred = out.sigmoid()
        else:
            loss = F.binary_cross_entropy_with_logits(out, kwargs['mask'].float())
            pred = out.sigmoid()

        return {
            "backward_loss": loss,
            f"pred_{self.output_type}": pred,
            "visual_loss": {"combined_loss": loss}
        }



# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------

import torch
import torch.nn as nn
from torch.nn import functional as F
from .GaussianSmoothing import GaussianSmoothing

class LaPlacianMs(nn.Module):
    def __init__(self,in_c,gauss_ker_size=3,scale=[2],drop_rate=0.2):
        super(LaPlacianMs, self).__init__()
        self.scale = scale
        self.gauss_ker_size = gauss_ker_size
        ## apply gaussian smoothing to input feature maps with 3 planes
        ## with kernel size K and sigma s
        self.smoothing = nn.ModuleDict()
        for s in self.scale:
            self.smoothing['scale-'+str(s)] = GaussianSmoothing(in_c, self.gauss_ker_size, s)
        self.conv_1x1 = nn.Sequential(nn.Conv2d(in_c*len(scale), in_c,
                                                kernel_size=1, stride=1,
                                                bias=False,groups=1),
                                                nn.BatchNorm2d(in_c),
                                                nn.ReLU(inplace=True),
                                                nn.Dropout(p=drop_rate)
        )
        # Official init from torch repo.
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def down(self,x,s):
        return F.interpolate(x,scale_factor=s,
                             mode='bilinear',
                             align_corners=False)
    def up (self,x, size):
        return F.interpolate(x,size=size,mode='bilinear',align_corners=False)

    def forward(self, x):
        for i, s in enumerate(self.scale):
            sm = self.smoothing['scale-'+str(s)](x)
            sm = self.down(sm,1/s)
            sm = self.up(sm,(x.shape[2],x.shape[3]))
            if i == 0:
                diff = x - sm
            else:
                diff = torch.cat((diff, x - sm), dim=1)
        return self.conv_1x1(diff)


import numpy as np
import timm
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
from .extractors.fph import FPH

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model
from IMDLBenCo.modules.extractors.high_frequency_feature_extraction import (
    FFTExtractor,
    DCTExtractor
)
from IMDLBenCo.modules.extractors.sobel import SobelFilter
from IMDLBenCo.modules.extractors.bayar_conv import BayerConv


class Resnet(BaseModel):
    def __init__(self,
                 input_head=None,
                 output_type='label',  # label or mask
                 backbone='resnet101',  # resnet50, reset101 or reset152
                 pretrained=True,  # if true, use imagenet-1k pretrained weight
                 image_size=256,  # to set mask size when output_type is mask
                 num_channels=3):
        super(Resnet, self).__init__()
        assert backbone in ['resnet50', 'resnet101', 'resnet152'], "Only resnet50, reset101 or reset152 are supported"
        self.model = timm.create_model(backbone, pretrained=pretrained)
        self.backbone = self.model.forward_features
        out_channels = self.model.num_features
        self.head = None
        self.output_type = output_type

        if output_type == 'label':
            self.head = nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Flatten(),
                nn.Linear(out_channels, 1)
            )
        elif output_type == 'mask':
            self.head = nn.Sequential(
                nn.Conv2d(out_channels, out_channels // 2, kernel_size=3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(out_channels // 2, out_channels // 4, kernel_size=3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(out_channels // 4, 1, kernel_size=1),
                nn.Upsample(size=(image_size, image_size), mode='bilinear', align_corners=False)
            )
        else:
            raise ValueError(f"Unsupported output_type: {output_type}")

        original_first_layer = list(self.model.children())[0]
        if input_head is not None:
            self.input_head = input_head
            new_first_layer = nn.Conv2d(num_channels + 3, original_first_layer.out_channels,
                                        kernel_size=original_first_layer.kernel_size,
                                        stride=original_first_layer.stride,
                                        padding=original_first_layer.padding, bias=False)
            new_first_layer.weight.data[:, :3, :, :] = original_first_layer.weight.data.clone()[:, :3, :, :]
            if num_channels > 0:
                new_first_layer.weight.data[:, 3:, :, :] = torch.nn.init.kaiming_normal_(
                    new_first_layer.weight[:, 3:, :, :])
            self.model.conv1 = new_first_layer
        else:
            self.input_head = None

    def forward(self, image, *args, **kwargs):
        if self.input_head is not None:
            feature = self.input_head(image)
            x = torch.cat([image, feature], dim=1)
        else:
            x = image
        out = self.head(self.backbone(x))

        if self.output_type == 'label':
            if len(out.shape) == 2:
                out = out.squeeze(dim=1)
            loss = F.binary_cross_entropy_with_logits(out, kwargs['label'].float())
            pred = out.sigmoid()
        else:
            loss = F.binary_cross_entropy_with_logits(out, kwargs['mask'].float())
            pred = out.sigmoid()
        out_dict = {
            "backward_loss": loss,
            f"pred_{self.output_type}": pred,
            "visual_loss": {
                "combined_loss": loss
            }
        }

        return out_dict


@register_model("Resnet50")
class Resnet50(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type, backbone='resnet50', pretrained=pretrained,
                         image_size=image_size, num_channels=0)


@register_model("Resnet101")
class Resnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type, backbone='resnet101', pretrained=pretrained,
                         image_size=image_size, num_channels=0)


@register_model("Resnet152")
class Resnet152(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type, backbone='resnet152', pretrained=pretrained,
                         image_size=image_size, num_channels=0)


@register_model("SobelResnet101")
class SobelResnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=SobelFilter(), output_type=output_type, backbone='resnet101', pretrained=pretrained,
                         image_size=image_size, num_channels=1)


@register_model("BayerResnet101")
class BayerResnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=BayerConv(), output_type=output_type, backbone='resnet101', pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("FFTResnet101")
class FFTResnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=FFTExtractor(), output_type=output_type, backbone='resnet101',
                         pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("DCTResnet101")
class DCTResnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=DCTExtractor(), output_type=output_type, backbone='resnet101',
                         pretrained=pretrained,
                         image_size=image_size, num_channels=3)


@register_model("QtResnet101")
class QtResnet101(Resnet):
    def __init__(self, output_type='label', pretrained=True, image_size=256):
        super().__init__(input_head=None, output_type=output_type,
                         backbone='resnet101', pretrained=pretrained,
                         image_size=image_size, num_channels=0)
        self.fph = FPH()

        # 定位主干中间模块
        self.stem = nn.Sequential(
            self.model.conv1,  # conv1
            self.model.bn1,
            self.model.act1,
            self.model.maxpool
        )
        # 主干后续部分
        self.blocks = nn.Sequential(
            self.model.layer1,
            self.model.layer2,
            self.model.layer3,
            self.model.layer4
        )

        # 通道融合（拼接后通道不一致）
        in_channels = self.model.layer1[0].conv1.in_channels + 256  # 通常是 64 + 256
        self.fusion_conv = nn.Conv2d(in_channels, self.model.layer1[0].conv1.in_channels, kernel_size=1)

    def forward(self, image, dct, qt, *args, **kwargs):
        dct = dct.squeeze(1).long()  # [B,1,H,W] -> [B,H,W]
        # FPH 特征（B, 256, H/8, W/8）
        x_aux = self.fph(dct, qt)

        # 主干前部分
        x = self.stem(image)

        # 若尺寸不一致，则插值对齐
        if x.shape[-2:] != x_aux.shape[-2:]:
            x_aux = F.interpolate(x_aux, size=x.shape[-2:], mode='bilinear', align_corners=False)

        # 拼接并降维融合
        x = torch.cat([x, x_aux], dim=1)
        x = self.fusion_conv(x)

        # 继续主干
        x = self.blocks(x)
        out = self.head(x)

        if self.output_type == 'label':
            if len(out.shape) == 2:
                out = out.squeeze(dim=1)
            loss = F.binary_cross_entropy_with_logits(out, kwargs['label'].float())
            pred = out.sigmoid()
        else:
            loss = F.binary_cross_entropy_with_logits(out, kwargs['mask'].float())
            pred = out.sigmoid()

        return {
            "backward_loss": loss,
            f"pred_{self.output_type}": pred,
            "visual_loss": {"combined_loss": loss}
        }



import sys

sys.path.append(".")
from ForensicHub.registry import DATASETS, MODELS, POSTFUNCS, TRANSFORMS, EVALUATORS

print(DATASETS._module_dict.keys())
print(MODELS._module_dict.keys())
print(POSTFUNCS._module_dict.keys())
print(TRANSFORMS._module_dict.keys())
print(EVALUATORS._module_dict.keys())



import torch
from torchvision import transforms
from .cospy_utils import data_augment
from .semantic_detector import SemanticDetector
from .artifact_detector import ArtifactDetector


# CO-SPY Calibrate Detector (Calibrate the integration of semantic and artifact detectors)
class CospyCalibrateDetector(torch.nn.Module):
    def __init__(self, semantic_weights_path, artifact_weights_path, num_classes=1):
        super(CospyCalibrateDetector, self).__init__()

        # Load the semantic detector
        self.sem = SemanticDetector()
        self.sem.load_weights(semantic_weights_path)

        # Load the artifact detector
        self.art = ArtifactDetector()
        self.art.load_weights(artifact_weights_path)

        # Freeze the two pre-trained models
        for param in self.sem.parameters():
            param.requires_grad = False
        for param in self.art.parameters():
            param.requires_grad = False

        # Classifier
        self.fc = torch.nn.Linear(2, num_classes)

        # Transformations inside the forward function
        # Including the normalization and resizing (only for the artifact detector)
        self.sem_transform = transforms.Compose([
            transforms.Normalize(self.sem.mean, self.sem.std)
        ])
        self.art_transform = transforms.Compose([
            transforms.Resize(self.art.cropSize, antialias=False),
            transforms.Normalize(self.art.mean, self.art.std)
        ])

        # Resolution
        self.loadSize = 224
        self.cropSize = 224

        # Data augmentation
        self.blur_prob = 0.0
        self.blur_sig = [0.0, 3.0]
        self.jpg_prob = 0.5
        self.jpg_method = ['cv2', 'pil']
        self.jpg_qual = list(range(70, 96))

        # Define the augmentation configuration
        self.aug_config = {
            "blur_prob": self.blur_prob,
            "blur_sig": self.blur_sig,
            "jpg_prob": self.jpg_prob,
            "jpg_method": self.jpg_method,
            "jpg_qual": self.jpg_qual,
        }

        # Pre-processing
        crop_func = transforms.RandomCrop(self.cropSize)
        flip_func = transforms.RandomHorizontalFlip()
        rz_func = transforms.Resize(self.loadSize)
        aug_func = transforms.Lambda(lambda x: data_augment(x, self.aug_config))

        self.train_transform = transforms.Compose([
            flip_func,
            aug_func,
            rz_func,
            crop_func,
            transforms.ToTensor(),
        ])

        self.test_transform = transforms.Compose([
            rz_func,
            crop_func,
            transforms.ToTensor(),
        ])

    def forward(self, x):
        x_sem = self.sem_transform(x)
        x_art = self.art_transform(x)
        pred_sem = self.sem(x_sem)
        pred_art = self.art(x_art)
        x = torch.cat([pred_sem, pred_art], dim=1)
        x = self.fc(x)
        return x

    def save_weights(self, weights_path):
        save_params = {"fc.weight": self.fc.weight.cpu(), "fc.bias": self.fc.bias.cpu()}
        torch.save(save_params, weights_path)

    def load_weights(self, weights_path):
        weights = torch.load(weights_path)
        self.fc.weight.data = weights["fc.weight"]
        self.fc.bias.data = weights["fc.bias"]



from PIL import Image
from torchvision import transforms
from os.path import join
import abc
import numpy as np
import torch
import torch.utils.data as data
import imageio
import os

class BaseData(data.Dataset):
    '''
        The dataset used for the IFDL dataset.
    '''
    def __init__(self, args):
        super(BaseData, self).__init__()
        self.crop_size = args.crop_size
        ## demo dataset:
        self.mani_data_dir = './data_dir'
        ## the full dataset:
        # self.mani_data_dir = './data'
        self.image_names = []
        self.image_class = []
        self.mask_names  = []

    def __getitem__(self, index):
        res = self.get_item(index)
        return res

    def __len__(self):
        return len(self.image_names)

    def generate_mask(self, mask):
        '''
            generate the corresponding binary mask.
        '''
        mask = mask.astype(np.float32) / 255
        mask[mask > 0.5] = 1
        mask[mask <= 0.5] = 0
        mask = np.expand_dims(mask, axis=0)
        mask = torch.from_numpy(mask)
        return mask

    def rgba2rgb(self, rgba, background=(255, 255, 255)):
        '''
            turn rgba to rgb.
        '''
        row, col, ch = rgba.shape
        rgb = np.zeros((row, col, 3), dtype='float32')
        r, g, b, a = rgba[:, :, 0], rgba[:, :, 1], rgba[:, :, 2], rgba[:, :, 3]
        a = np.asarray(a, dtype='float32') / 255.0

        R, G, B = background
        rgb[:, :, 0] = r * a + (1.0 - a) * R
        rgb[:, :, 1] = g * a + (1.0 - a) * G
        rgb[:, :, 2] = b * a + (1.0 - a) * B
        return np.asarray(rgb, dtype='uint8') # the output value is uint8 that belongs to [0,255]

    def get_image(self, image_name):
        '''
            return the image with the tensor.
        '''
        image = imageio.imread(image_name)
        if len(image.shape) == 2:
            image = imageio.imread(image_name, as_gray=False, pilmode="RGB")
        if image.shape[-1] == 4:
            image = self.rgba2rgb(image)
        image = torch.from_numpy(image.astype(np.float32) / 255)
        return image.permute(2, 0, 1)

    def get_mask(self, mask_name):
        '''
            return the binary mask.
        '''   
        mask = Image.open(mask_name).convert('L')
        mask = mask.resize(self.crop_size, resample=Image.BICUBIC)
        mask = np.asarray(mask)
        mask = self.generate_mask(mask) 
        return mask

    @abc.abstractmethod
    def get_item(self, index):
        '''
            blur
            image = Image.fromarray(image)
            image = image.filter(ImageFilter.GaussianBlur(radius=7))
            image = np.asarray(image)

            resize
            image = Image.fromarray(image)
            image = image.resize((int(image.width*0.25), int(image.height*0.25)), resample=Image.BILINEAR)
            image = np.asarray(image)

            noise
            import skimage
            image = skimage.util.random_noise(image/255., mode='gaussian', mean=0, var=15/255) * 255

            jpeg compression
            im = Image.open(image_name)
            temp_name = './temp/' + image_name.split('/')[-1][:-3] + 'jpg'
            im.save(temp_name, 'JPEG', quality=50)
            image = Image.open(temp_name)
            image = np.asarray(image)
        '''
        pass

class ValColumbia(BaseData):
    def __init__(self, args):
        super(ValColumbia, self).__init__(args)
        ddir = os.path.join(self.mani_data_dir, 'columbia')
        with open(join(ddir, 'vallist.txt')) as f:
            contents = f.readlines()
            for content in contents:
                _ = os.path.join(ddir, '4cam_splc', content.strip())
                self.image_names.append(_)
        self.image_class = [1] * len(self.image_names)

    def get_item(self, index):
        image_name = self.image_names[index]
        cls = self.image_class[index]

        # image
        image = self.get_image(image_name)

        # mask
        if '4cam_splc' in image_name:
            mask_name = image_name.replace('4cam_splc', 'mask').replace('.tif', '.jpg')
            mask = self.get_mask(mask_name)
        else:
            mask = np.zeros((1, 256, 256), dtype='float32')

        return image, mask, cls, image_name

class ValCoverage(BaseData):
    def __init__(self, args):
        super(ValCoverage, self).__init__(args)
        ddir = os.path.join(self.mani_data_dir, 'Coverage')
        with open(join(ddir, 'fake.txt')) as f:
            contents = f.readlines()
            for content in contents:
                _ = os.path.join(ddir, 'image', content.strip())
                self.image_names.append(_)
        self.image_class = [2] * len(self.image_names)

    def get_item(self, index):
        image_name = self.image_names[index]
        cls = self.image_class[index]

        ## read image.
        image = self.get_image(image_name)

        # mask
        mask_name = image_name.replace('image', 'mask').replace('t.tif', 'forged.tif')
        mask = self.get_mask(mask_name)

        return image, mask, cls, image_name

class ValCasia(BaseData):
    def __init__(self, args):
        super(ValCasia, self).__init__(args)
        ddir = os.path.join(self.mani_data_dir, 'CASIA/CASIA1')
        with open(join(ddir, 'fake.txt')) as f:
            contents = f.readlines()
            for content in contents:
                tag = content.split('/')[-1].split('_')[1]
                if tag == 'D':
                    self.image_class.append(1)
                elif tag == 'S':
                    self.image_class.append(2)
                else:
                    raise Exception('unknown class: {}'.format(content))
                self.image_names.append(os.path.join(ddir, 'fake', content.strip()))

        ddir = os.path.join(self.mani_data_dir, 'CASIA/CASIA2')
        with open(join(ddir, 'fake.txt')) as f:
            contents = f.readlines()
            for content in contents:
                tag = content.split('/')[-1].split('_')[1]
                if tag == 'D':
                    self.image_class.append(1)
                elif tag == 'S':
                    self.image_class.append(2)
                else:
                    raise Exception('unknown class: {}'.format(content))
                self.image_names.append(os.path.join(ddir, 'fake', content.strip()))

    def get_item(self, index):
        image_name = self.image_names[index]
        cls = self.image_class[index]

        # image
        image = self.get_image(image_name)

        # mask
        if '.jpg' in image_name:
            mask_name = image_name.replace('fake', 'mask').replace('.jpg', '_gt.png')
        else:
            mask_name = image_name.replace('fake', 'mask').replace('.tif', '_gt.png')
        mask = self.get_mask(mask_name)

        return image, mask, cls, image_name

class ValNIST16(BaseData):
    def __init__(self, args):
        super(ValNIST16, self).__init__(args)
        ddir = os.path.join(self.mani_data_dir, 'NIST16')
        file_name = 'alllist.txt'
        with open(join(ddir, file_name)) as f:
            contents = f.readlines()
            for content in contents:
                image_name, mask_name = content.split(' ')
                self.image_names.append(join(ddir, image_name))
                self.mask_names.append(join(ddir, mask_name.strip()))

    def get_item(self, index):
        image_name = self.image_names[index]
        mask_name = self.mask_names[index]

        if 'splice' in mask_name:
            cls = 1
        elif 'manipulation' in mask_name:
            cls = 2
        elif 'remove' in mask_name:
            cls = 3
        else:
            cls = 0

        # image
        image = self.get_image(image_name)
        if image.size()[2]*image.size()[1] >= 1000*1000:
            image = imageio.imread(image_name)
            if image.shape[-1] == 4:
                image = self.rgba2rgb(image)
            image = Image.fromarray(image)
            image = image.resize((1000, 1000), resample=Image.BICUBIC)
            image = np.asarray(image)
            image = torch.from_numpy(image.astype(np.float32) / 255)
            image = image.permute(2, 0, 1)

        # mask
        mask = self.get_mask(mask_name)
        mask = torch.abs(mask - 1)

        return image, mask, cls, image_name

class ValIMD2020(BaseData):
    def __init__(self, args):
        super(ValIMD2020, self).__init__(args)
        ddir = os.path.join(self.mani_data_dir, 'IMD2020')
        file_name = 'fake.txt'
        with open(join(ddir, file_name)) as f:
            contents = f.readlines()
            for content in contents:
                image_name = content.strip()
                if '.jpg' in image_name:
                    mask_name = image_name.replace('.jpg', '_mask.png')
                else:
                    mask_name = image_name.replace('.png', '_mask.png')
                self.image_names.append(join(ddir, 'fake_img', image_name))
                self.mask_names.append(join(ddir, 'mask', mask_name))
        self.image_class = [2] * len(self.image_names)

    def get_item(self, index):
        image_name = self.image_names[index]
        mask_name = self.mask_names[index]
        cls = self.image_class[index]
        try:
            image = self.get_image(image_name)
        except:
            print(f"Fail at {image_name}.")
        mask = self.get_mask(mask_name)

        return image, mask, cls, image_name


"""
Restormer architecture for ADCD-Net
Based on "Restormer: Efficient Transformer for High-Resolution Image Restoration"
https://arxiv.org/abs/2111.09881

Modified for document tampering detection with DCT feature integration.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint
import numbers


def to_3d(x):
    """Convert (B, C, H, W) to (B, H*W, C)"""
    b, c, h, w = x.shape
    return x.permute(0, 2, 3, 1).reshape(b, h * w, c)


def to_4d(x, h, w):
    """Convert (B, H*W, C) to (B, C, H, W)"""
    b, _, c = x.shape
    return x.reshape(b, h, w, c).permute(0, 3, 1, 2)


class BiasFree_LayerNorm(nn.Module):
    def __init__(self, normalized_shape):
        super(BiasFree_LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        normalized_shape = torch.Size(normalized_shape)

        assert len(normalized_shape) == 1

        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.normalized_shape = normalized_shape

    def forward(self, x):
        sigma = x.var(-1, keepdim=True, unbiased=False)
        return x / torch.sqrt(sigma + 1e-5) * self.weight


class WithBias_LayerNorm(nn.Module):
    def __init__(self, normalized_shape):
        super(WithBias_LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        normalized_shape = torch.Size(normalized_shape)

        assert len(normalized_shape) == 1

        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.normalized_shape = normalized_shape

    def forward(self, x):
        mu = x.mean(-1, keepdim=True)
        sigma = x.var(-1, keepdim=True, unbiased=False)
        return (x - mu) / torch.sqrt(sigma + 1e-5) * self.weight + self.bias


class LayerNorm(nn.Module):
    def __init__(self, dim, LayerNorm_type='WithBias'):
        super(LayerNorm, self).__init__()
        if LayerNorm_type == 'BiasFree':
            self.body = BiasFree_LayerNorm(dim)
        else:
            self.body = WithBias_LayerNorm(dim)

    def forward(self, x):
        h, w = x.shape[-2:]
        return to_4d(self.body(to_3d(x)), h, w)


class FeedForward(nn.Module):
    def __init__(self, dim, ffn_expansion_factor=2.66, bias=False):
        super(FeedForward, self).__init__()

        hidden_features = int(dim * ffn_expansion_factor)

        self.project_in = nn.Conv2d(dim, hidden_features * 2, kernel_size=1, bias=bias)

        self.dwconv = nn.Conv2d(hidden_features * 2, hidden_features * 2, kernel_size=3,
                                stride=1, padding=1, groups=hidden_features * 2, bias=bias)

        self.project_out = nn.Conv2d(hidden_features, dim, kernel_size=1, bias=bias)

    def forward(self, x):
        x = self.project_in(x)
        x1, x2 = self.dwconv(x).chunk(2, dim=1)
        x = F.gelu(x1) * x2
        x = self.project_out(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, bias=False):
        super(Attention, self).__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Conv2d(dim, dim * 3, kernel_size=1, bias=bias)
        self.qkv_dwconv = nn.Conv2d(dim * 3, dim * 3, kernel_size=3, stride=1, padding=1,
                                    groups=dim * 3, bias=bias)
        self.project_out = nn.Conv2d(dim, dim, kernel_size=1, bias=bias)

    def forward(self, x):
        b, c, h, w = x.shape

        qkv = self.qkv_dwconv(self.qkv(x))
        q, k, v = qkv.chunk(3, dim=1)

        # Reshape: (B, C, H, W) -> (B, num_heads, C//num_heads, H*W)
        head_dim = c // self.num_heads
        q = q.reshape(b, self.num_heads, head_dim, h * w)
        k = k.reshape(b, self.num_heads, head_dim, h * w)
        v = v.reshape(b, self.num_heads, head_dim, h * w)

        q = F.normalize(q, dim=-1)
        k = F.normalize(k, dim=-1)

        attn = (q @ k.transpose(-2, -1)) * self.temperature
        attn = attn.softmax(dim=-1)

        out = (attn @ v)

        # Reshape back: (B, num_heads, C//num_heads, H*W) -> (B, C, H, W)
        out = out.reshape(b, c, h, w)

        out = self.project_out(out)
        return out


class TransformerBlock(nn.Module):
    def __init__(self, dim, num_heads=8, ffn_expansion_factor=2.66, bias=False, LayerNorm_type='WithBias'):
        super(TransformerBlock, self).__init__()

        self.norm1 = LayerNorm(dim, LayerNorm_type)
        self.attn = Attention(dim, num_heads, bias)
        self.norm2 = LayerNorm(dim, LayerNorm_type)
        self.ffn = FeedForward(dim, ffn_expansion_factor, bias)

    def forward(self, x):
        x = x + self.attn(self.norm1(x))
        x = x + self.ffn(self.norm2(x))
        return x


class OverlapPatchEmbed(nn.Module):
    def __init__(self, in_c=3, embed_dim=48, bias=False):
        super(OverlapPatchEmbed, self).__init__()

        self.proj = nn.Conv2d(in_c, embed_dim, kernel_size=3, stride=1, padding=1, bias=bias)

    def forward(self, x):
        x = self.proj(x)
        return x


class Downsample(nn.Module):
    def __init__(self, n_feat):
        super(Downsample, self).__init__()

        self.body = nn.Sequential(
            nn.Conv2d(n_feat, n_feat // 2, kernel_size=3, stride=1, padding=1, bias=False),
            nn.PixelUnshuffle(2)
        )

    def forward(self, x):
        return self.body(x)


class Upsample(nn.Module):
    def __init__(self, n_feat):
        super(Upsample, self).__init__()

        self.body = nn.Sequential(
            nn.Conv2d(n_feat, n_feat * 2, kernel_size=3, stride=1, padding=1, bias=False),
            nn.PixelShuffle(2)
        )

    def forward(self, x):
        return self.body(x)


class DCTFusion(nn.Module):
    """Fuse DCT features with RGB features using alignment score."""

    def __init__(self, rgb_dim, dct_dim):
        super().__init__()
        self.rgb_proj = nn.Conv2d(rgb_dim, rgb_dim, 1)
        self.dct_proj = nn.Conv2d(dct_dim, rgb_dim, 1)
        self.fusion = nn.Sequential(
            nn.Conv2d(rgb_dim * 2, rgb_dim, 3, 1, 1),
            nn.BatchNorm2d(rgb_dim),
            nn.ReLU(inplace=True)
        )

    def forward(self, rgb_feat, dct_feat, align_score):
        """
        Args:
            rgb_feat: RGB features (B, C_rgb, H, W)
            dct_feat: DCT features (B, C_dct, H', W')
            align_score: Alignment score (B,) with values 0-1
        """
        # Resize DCT features to match RGB features
        if dct_feat.shape[-2:] != rgb_feat.shape[-2:]:
            dct_feat = F.interpolate(dct_feat, size=rgb_feat.shape[-2:], mode='bilinear', align_corners=False)

        rgb_proj = self.rgb_proj(rgb_feat)
        dct_proj = self.dct_proj(dct_feat)

        # Modulate DCT contribution by alignment score
        align_score = align_score.view(-1, 1, 1, 1)
        dct_modulated = dct_proj * align_score

        fused = self.fusion(torch.cat([rgb_proj, dct_modulated], dim=1))
        return fused + rgb_feat


class Restormer(nn.Module):
    """
    Restormer-based encoder-decoder for ADCD-Net.

    Modified to output multi-scale features for:
    - Localization branch
    - Content/foreground disentanglement
    - Pristine prototype estimation
    """

    def __init__(
            self,
            inp_channels=6,  # RGB (3) + coord (2) + placeholder (1)
            out_channels=96,
            dim=48,
            num_blocks=[4, 6, 6, 8],
            num_refinement_blocks=4,
            heads=[1, 2, 4, 8],
            ffn_expansion_factor=2.66,
            bias=False,
            LayerNorm_type='WithBias',
            dct_dims=[128, 64, 48, 48],  # DCT feature dimensions at each scale
            use_checkpoint=False  # Gradient checkpointing for memory saving
    ):
        super(Restormer, self).__init__()

        self.use_checkpoint = use_checkpoint
        self.patch_embed = OverlapPatchEmbed(inp_channels, dim)

        # Encoder
        self.encoder_level1 = nn.Sequential(*[
            TransformerBlock(dim=dim, num_heads=heads[0], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[0])
        ])

        self.down1_2 = Downsample(dim)
        self.encoder_level2 = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 1), num_heads=heads[1], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[1])
        ])

        self.down2_3 = Downsample(int(dim * 2 ** 1))
        self.encoder_level3 = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 2), num_heads=heads[2], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[2])
        ])

        self.down3_4 = Downsample(int(dim * 2 ** 2))
        self.latent = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 3), num_heads=heads[3], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[3])
        ])

        # DCT fusion modules
        self.dct_fusion1 = DCTFusion(dim, dct_dims[0])
        self.dct_fusion2 = DCTFusion(int(dim * 2), dct_dims[1])
        self.dct_fusion3 = DCTFusion(int(dim * 4), dct_dims[2])
        self.dct_fusion4 = DCTFusion(int(dim * 8), dct_dims[3])

        # Decoder
        self.up4_3 = Upsample(int(dim * 2 ** 3))
        self.reduce_chan_level3 = nn.Conv2d(int(dim * 2 ** 3), int(dim * 2 ** 2), kernel_size=1, bias=bias)
        self.decoder_level3 = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 2), num_heads=heads[2], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[2])
        ])

        self.up3_2 = Upsample(int(dim * 2 ** 2))
        self.reduce_chan_level2 = nn.Conv2d(int(dim * 2 ** 2), int(dim * 2 ** 1), kernel_size=1, bias=bias)
        self.decoder_level2 = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 1), num_heads=heads[1], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[1])
        ])

        self.up2_1 = Upsample(int(dim * 2 ** 1))
        self.decoder_level1 = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 1), num_heads=heads[0], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_blocks[0])
        ])

        self.refinement = nn.Sequential(*[
            TransformerBlock(dim=int(dim * 2 ** 1), num_heads=heads[0], ffn_expansion_factor=ffn_expansion_factor,
                             bias=bias, LayerNorm_type=LayerNorm_type)
            for i in range(num_refinement_blocks)
        ])

        self.output = nn.Conv2d(int(dim * 2 ** 1), out_channels, kernel_size=3, stride=1, padding=1, bias=bias)

        # Feature extractors for disentanglement
        self.cnt_head = nn.Conv2d(int(dim * 2 ** 1), out_channels, 3, 1, 1)
        self.frg_head = nn.Conv2d(int(dim * 2 ** 1), out_channels, 3, 1, 1)

    def _run_block(self, block, x):
        """Run a block with optional gradient checkpointing."""
        if self.use_checkpoint and self.training:
            return checkpoint(block, x, use_reentrant=False)
        return block(x)

    def forward(self, img, ms_dct_feats=None, dct_align_score=None):
        """
        Args:
            img: Input image with coordinates (B, 6, H, W)
            ms_dct_feats: Multi-scale DCT features from FPH
            dct_align_score: Alignment score from FPH (B,)

        Returns:
            feat: Main localization feature (B, out_channels, H, W)
            cnt_feats: Content features for reconstruction
            frg_feats: Foreground features for reconstruction
            pp_feats: Pristine prototype features at multiple scales
        """
        inp_enc_level1 = self.patch_embed(img)

        # Encoder with DCT fusion (with optional checkpointing)
        out_enc_level1 = self._run_block(self.encoder_level1, inp_enc_level1)
        if ms_dct_feats is not None and dct_align_score is not None:
            out_enc_level1 = self.dct_fusion1(out_enc_level1, ms_dct_feats[0], dct_align_score)

        inp_enc_level2 = self.down1_2(out_enc_level1)
        out_enc_level2 = self._run_block(self.encoder_level2, inp_enc_level2)
        if ms_dct_feats is not None and dct_align_score is not None:
            out_enc_level2 = self.dct_fusion2(out_enc_level2, ms_dct_feats[1], dct_align_score)

        inp_enc_level3 = self.down2_3(out_enc_level2)
        out_enc_level3 = self._run_block(self.encoder_level3, inp_enc_level3)
        if ms_dct_feats is not None and dct_align_score is not None:
            out_enc_level3 = self.dct_fusion3(out_enc_level3, ms_dct_feats[2], dct_align_score)

        inp_enc_level4 = self.down3_4(out_enc_level3)
        latent = self._run_block(self.latent, inp_enc_level4)
        if ms_dct_feats is not None and dct_align_score is not None:
            latent = self.dct_fusion4(latent, ms_dct_feats[3], dct_align_score)

        # Decoder (with optional checkpointing)
        inp_dec_level3 = self.up4_3(latent)
        inp_dec_level3 = torch.cat([inp_dec_level3, out_enc_level3], 1)
        inp_dec_level3 = self.reduce_chan_level3(inp_dec_level3)
        out_dec_level3 = self._run_block(self.decoder_level3, inp_dec_level3)

        inp_dec_level2 = self.up3_2(out_dec_level3)
        inp_dec_level2 = torch.cat([inp_dec_level2, out_enc_level2], 1)
        inp_dec_level2 = self.reduce_chan_level2(inp_dec_level2)
        out_dec_level2 = self._run_block(self.decoder_level2, inp_dec_level2)

        inp_dec_level1 = self.up2_1(out_dec_level2)
        inp_dec_level1 = torch.cat([inp_dec_level1, out_enc_level1], 1)
        out_dec_level1 = self._run_block(self.decoder_level1, inp_dec_level1)

        out_dec_level1 = self._run_block(self.refinement, out_dec_level1)

        feat = self.output(out_dec_level1)

        # Content and foreground features for reconstruction
        cnt_feats = self.cnt_head(out_dec_level1)
        frg_feats = self.frg_head(out_dec_level1)

        # Pristine prototype features at multiple scales
        pp_feats = [
            F.interpolate(out_dec_level1, scale_factor=0.5, mode='bilinear', align_corners=False),
            out_dec_level2,
            out_dec_level3,
            latent
        ]

        return feat, cnt_feats, frg_feats, pp_feats


class RestormerDecoder(nn.Module):
    """Decoder-only Restormer for reconstruction branch."""

    def __init__(self, in_channels=96, out_channels=4, dim=48):
        super().__init__()

        self.proj = nn.Conv2d(in_channels * 2, dim * 2, 1)

        self.decoder = nn.Sequential(
            TransformerBlock(dim * 2, num_heads=2),
            TransformerBlock(dim * 2, num_heads=2),
            Upsample(dim * 2),
            TransformerBlock(dim, num_heads=1),
            TransformerBlock(dim, num_heads=1),
        )

        self.output = nn.Conv2d(dim, out_channels, 3, 1, 1)

    def forward(self, cnt_feats, frg_feats, is_shuffle=False):
        """
        Reconstruct image from content and foreground features.

        Args:
            cnt_feats: Content features (B, C, H, W)
            frg_feats: Foreground features (B, C, H, W)
            is_shuffle: Whether to shuffle foreground features for contrastive learning
        """
        if is_shuffle:
            # Shuffle foreground features across batch
            idx = torch.randperm(frg_feats.size(0))
            frg_feats = frg_feats[idx]

        x = torch.cat([cnt_feats, frg_feats], dim=1)
        x = self.proj(x)
        x = self.decoder(x)
        return self.output(x)


def get_restormer(model_name='full_model', out_channels=96, **kwargs):
    """
    Factory function to create Restormer models.

    Args:
        model_name: 'full_model' for encoder-decoder, 'decoder_only' for reconstruction
        out_channels: Number of output channels
    """
    if model_name == 'full_model':
        return Restormer(out_channels=out_channels, **kwargs)
    elif model_name == 'decoder_only':
        return RestormerDecoder(out_channels=out_channels, **kwargs)
    else:
        raise ValueError(f"Unknown model name: {model_name}")


if __name__ == "__main__":
    # Test full model
    model = get_restormer('full_model', out_channels=96)
    img = torch.randn(2, 6, 256, 256)
    ms_dct_feats = [
        torch.randn(2, 128, 32, 32),
        torch.randn(2, 64, 16, 16),
        torch.randn(2, 48, 8, 8),
        torch.randn(2, 48, 4, 4),
    ]
    align_score = torch.rand(2)

    feat, cnt_feats, frg_feats, pp_feats = model(img, ms_dct_feats, align_score)
    print(f"Main feature: {feat.shape}")
    print(f"Content features: {cnt_feats.shape}")
    print(f"Foreground features: {frg_feats.shape}")
    for i, f in enumerate(pp_feats):
        print(f"PP feature {i}: {f.shape}")

    # Test decoder
    decoder = get_restormer('decoder_only', out_channels=4)
    rec = decoder(cnt_feats, frg_feats, is_shuffle=True)
    print(f"Reconstructed: {rec.shape}")



import torch
import torchvision
import torch.nn as nn


class FocalLoss(nn.Module):
    def __init__(self,
                 alpha=0.25,
                 gamma=2,
                 reduction='mean', ):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        self.crit = nn.BCELoss(reduction='none')

    def forward(self, logits, label):
        # compute loss
        logits = logits.float()  # use fp32 if logits is fp16
        with torch.no_grad():
            alpha = torch.empty_like(logits).fill_(1 - self.alpha)
            alpha[label == 1] = self.alpha

        probs = torch.sigmoid(logits)
        pt = torch.where(label == 1, probs, 1 - probs)
        ce_loss = self.crit(logits, label)
        loss = (alpha * torch.pow(1 - pt, self.gamma) * ce_loss)
        if self.reduction == 'mean':
            loss = loss.mean()
        if self.reduction == 'sum':
            loss = loss.sum()
        return loss

class Loss_Function(nn.Module):
    def __init__(self):
        super(Loss_Function, self).__init__()
        #self.bceloss = nn.BCELoss()
        #self.focalloss = FocalLoss()
        self.ce = nn.CrossEntropyLoss()
    def forward(self, logits, label):
        #bce_loss = self.bceloss(logits,label)
        #focal_loss = self.focalloss(logits,label)
        ce_loss = self.ce(logits,label)
        return ce_loss


from .datasets import *
from .models import *
from .transforms import *

import types

from IMDLBenCo.model_zoo import __all__
from IMDLBenCo.model_zoo import *
from ForensicHub import MODELS, POSTFUNCS

# 确保模型类注册到框架
for model_name in __all__:
    # 获取对象
    obj = globals().get(model_name)

    if obj is None:
        continue  # 没找到，跳过

    if isinstance(obj, type):
        # 是类，注册到 MODELS
        if obj.__module__.startswith('IMDLBenCo.model_zoo'):
            MODELS.register_module(model_name, force=True, module=obj)
    elif isinstance(obj, types.FunctionType):
        # 是函数，注册到 POSTFUNCS
        POSTFUNCS.register_module(model_name, force=True, module=obj)
    else:
        # 其他情况，可以根据需要提示或跳过
        pass



import torch
import torch.nn as nn
from typing import Union, List, Dict, Any, cast
import torchvision
import torch.nn.functional as F





class VGG(torch.nn.Module):
    def __init__(self, arch_type, pretrained, progress):
        super().__init__()
        
        self.layer1 = torch.nn.Sequential()
        self.layer2 = torch.nn.Sequential()
        self.layer3 = torch.nn.Sequential()
        self.layer4 = torch.nn.Sequential()
        self.layer5 = torch.nn.Sequential()

        if arch_type == 'vgg11':
            official_vgg = torchvision.models.vgg11(pretrained=pretrained, progress=progress)
            blocks = [  [0,2], [2,5], [5,10], [10,15], [15,20] ]
            last_idx = 20
        elif arch_type == 'vgg19':
            official_vgg = torchvision.models.vgg19(pretrained=pretrained, progress=progress)
            blocks = [  [0,4], [4,9], [9,18], [18,27], [27,36] ]
            last_idx = 36
        else:
            raise NotImplementedError
        
        
        for x in range( *blocks[0] ):
            self.layer1.add_module(str(x), official_vgg.features[x])
        for x in range( *blocks[1] ):
            self.layer2.add_module(str(x), official_vgg.features[x])
        for x in range( *blocks[2] ):
            self.layer3.add_module(str(x), official_vgg.features[x])
        for x in range( *blocks[3] ):
            self.layer4.add_module(str(x), official_vgg.features[x])
        for x in range( *blocks[4] ):
            self.layer5.add_module(str(x), official_vgg.features[x])
            
        self.max_pool = official_vgg.features[last_idx]
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        
        self.fc1 = official_vgg.classifier[0]
        self.fc2 = official_vgg.classifier[3]
        self.fc3 = official_vgg.classifier[6]
        self.dropout = nn.Dropout()
        
        
    def forward(self, x):
        out = {}
        
        x = self.layer1(x)
        out['f0'] = x
        
        x = self.layer2(x)
        out['f1'] = x
        
        x = self.layer3(x)
        out['f2'] = x
        
        x = self.layer4(x)
        out['f3'] = x
        
        x = self.layer5(x)
        out['f4'] = x
        
        x = self.max_pool(x)
        x = self.avgpool(x)
        x = x.view(-1,512*7*7) 
        
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout(x) 
        x = self.fc2(x)
        x = F.relu(x)
        out['penultimate'] = x 
        x = self.dropout(x) 
        x = self.fc3(x)
        out['logits'] = x 

        return out










def vgg11(pretrained=False, progress=True):
    r"""VGG 11-layer model (configuration "A") from
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return VGG('vgg11', pretrained, progress)



def vgg19(pretrained=False, progress=True):
    r"""VGG 19-layer model (configuration "E")
    `"Very Deep Convolutional Networks For Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return VGG('vgg19', pretrained, progress)







import torch
import torch.nn as nn
from typing import Dict, Any, Optional


class BaseModel(nn.Module):
    """Base class for all models in ForensicHub.
    
    This class defines the basic interface and common functionality
    for all models. The model should take at least an image as input
    and return a dictionary containing the necessary outputs.
    """

    def __init__(self):
        super().__init__()

    def forward(self, **kwargs) -> Dict[str, Any]:
        """Forward pass of the model.
        
        Args:
            **kwargs (Dict[str, Any]): Dictionary containing input data.
                May contain (Consistent with the input from the dataset):
                    - 'image': Input image tensor
                    - 'mask': Ground truth mask
                    - 'label': Ground truth label
                    - Other model-specific inputs
                    
        Returns:
            Dict[str, Any]: Dictionary containing model outputs.
                Must contain at least:
                    - 'backward_loss': Backward loss value
                May contain:
                    - 'pred_mask': Predicted mask
                    - 'pred_label': Predicted label
                Visualize (Optional):
                    - 'visual_loss': Automatically visualize with the key-value pairs 
                    eg. {
                        "predict_loss": predict_loss,
                        "edge_loss": edge_loss,
                        "combined_loss": combined_loss
                    }
                    - 'visual_image': Automatically visualize with the key-value pairs
                    eg. {
                        "mask": mask,
                        "pred_mask": pred_mask,
                        "pred_label": pred_label
                    }
                    - Other model-specific outputs

                return eg.
                {
                    "backward_loss": combined_loss,

                    # optional below
                    "pred_mask": mask_pred,
                    "visual_loss": {
                        "combined_loss": combined_loss
                    },
                    "visual_image": {
                        "pred_mask": mask_pred,
                    }
                }
        """
        raise NotImplementedError("Subclasses must implement forward method")

    def get_prediction(self, data_dict: Dict[str, Any]) -> Dict[str, Any]:
        """Get model prediction without computing loss.
        
        This is typically used during inference.
        
        Args:
            data_dict (Dict[str, Any]): Dictionary containing input data.
                Must contain at least:
                    - 'image': Input image tensor
                    
        Returns:
            Dict[str, Any]: Dictionary containing model outputs.
                Must contain at least:
                    - 'pred': Model prediction
        """
        with torch.no_grad():
            return self.forward(**data_dict)

    def compute_loss(self, data_dict: Dict[str, Any], output_dict: Dict[str, Any]) -> torch.Tensor:
        """Compute loss for training.
        
        Args:
            data_dict (Dict[str, Any]): Dictionary containing input data.
            output_dict (Dict[str, Any]): Dictionary containing model outputs.
                
        Returns:
            torch.Tensor: Loss value
        """
        raise NotImplementedError("Subclasses must implement compute_loss method")

    def get_metrics(self, data_dict: Dict[str, Any], output_dict: Dict[str, Any]) -> Dict[str, float]:
        """Compute metrics for evaluation.
        
        Args:
            data_dict (Dict[str, Any]): Dictionary containing input data.
            output_dict (Dict[str, Any]): Dictionary containing model outputs.
                
        Returns:
            Dict[str, float]: Dictionary containing metric names and values
        """
        raise NotImplementedError("Subclasses must implement get_metrics method")



from .backbones.resnet import *
from .backbones.segformer import *
from .backbones.convnext import *
from .backbones.swintransformer import *
from .backbones.efficientnet import *
from .backbones.xception import *
from .backbones.unet import *
from .backbones.mobilenet import *
from .backbones.vit import *
from .backbones.densenet import *
from .backbones.qwen3_vl_llm import *
from .wrapper.mask2labelwrapper import Mask2LabelWrapper
from .wrapper.crossdatasetwrapper import CrossDataset



# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------

import math
import numbers
import torch
from torch import nn
from torch.nn import functional as F

class GaussianSmoothing(nn.Module):
    """
    Apply gaussian smoothing on a
    1d, 2d or 3d tensor. Filtering is performed seperately for each channel
    in the input using a depthwise convolution.
    Arguments:
    channels (int, sequence): Number of channels of the input tensors. Output will
    have this number of channels as well.
    kernel_size (int, sequence): Size of the gaussian kernel.
    sigma (float, sequence): Standard deviation of the gaussian kernel.
    dim (int, optional): The number of dimensions of the data.
    Default value is 2 (spatial).
    """
    def __init__(self, channels, kernel_size, sigma, dim=2):
        super(GaussianSmoothing, self).__init__()
        if isinstance(kernel_size, numbers.Number):
            kernel_size = [kernel_size] * dim
        if isinstance(sigma, numbers.Number):
            sigma = [sigma] * dim

        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [
                torch.arange(size, dtype=torch.float32)
                for size in kernel_size
            ], indexing='ij'
        )
        for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
            mean = (size - 1) / 2
            kernel *= 1 / (std * math.sqrt(2 * math.pi)) * \
                      torch.exp(-((mgrid - mean) / std) ** 2 / 2)
        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))
        self.register_buffer('weight', kernel)
        self.groups = channels

        if dim == 1:
            self.conv = F.conv1d
        elif dim == 2:
            self.conv = F.conv2d
        elif dim == 3:
            self.conv = F.conv3d
        else:
            raise RuntimeError(
                'Only 1, 2 and 3 dimensions are supported. Received {}.'.format(dim)
            )

    def forward(self, input):
        """
        Apply gaussian filter to input.
        Arguments:
        input (torch.Tensor): Input to apply gaussian filter on.
        Returns:
        filtered (torch.Tensor): Filtered output.
        """
        return self.conv(input, weight=self.weight, groups=self.groups)


# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------
import os
import time
import torch
import torch.nn as nn
from tqdm import tqdm

device = torch.device('cuda:0')
device_ids = [0]

class IsolatingLossFunction(torch.nn.Module):
	def __init__(self, c, R, p=2, threshold_val=1.85):
		super().__init__()
		self.c = c.clone().detach() # Center of the hypershpere, c ∈ ℝ^d (d-dimensional real-valued vector)
		self.R = R.clone().detach() # Radius of the hypersphere, R ∈ ℝ^1 (Real-valued)
		self.p = p                  # norm value (p-norm), p ∈ ℝ^1 (Default 2)
		self.margin_natu = (0.15)*self.R    
		self.margin_mani = (2.5)*self.R
		self.threshold   = threshold_val*self.R

		print('\n')
		print('$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$')
		print('$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$')
		print(f'The Radius manipul is {self.margin_natu}.')
		print(f'The Radius expansn is {self.margin_mani}.')
		print(f'The Radius threshold is {self.threshold}.')
		print('$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$')
		print('$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$')
		print('\n')
		self.pdist = torch.nn.PairwiseDistance(p=self.p) # Creating a Pairwise Distance object
		self.dis_curBatch = 0
		self.dis = 0

	def forward(self, model_output, label, threshold_new=None, update_flag=None):
		'''output the distance mask and compute the loss.'''
		bs, feat_dim, w, h = model_output.size()
		model_output = model_output.permute(0,2,3,1)
		model_output = torch.reshape(model_output, (-1, feat_dim))
		dist = self.pdist(model_output, self.c)
		self.dist = dist
		pred_mask = torch.gt(self.dist, self.threshold).to(torch.float32)
		pred_mask = torch.reshape(pred_mask, (bs,w,h,1)).permute(0,3,1,2)
		self.dist = torch.reshape(self.dist, (bs,w,h,1)).permute(0,3,1,2)
		self.dis_curBatch = pred_mask.to(device).to(torch.float32)

		label = torch.reshape(label, (bs*w*h,1))
		label_sum = label.sum().item()

		label_nat  = torch.eq(label,0)
		label_mani = torch.eq(label,1)
		assert dist.size() == label_nat[:,0].size() 
		assert dist.size() == label_mani[:,0].size()

		label_nat_sum  = label_nat.sum().item()
		label_mani_sum = label_mani.sum().item()

		dist_nat  = torch.masked_select(dist, label_nat[:,0])
		dist_mani = torch.max(torch.tensor(0).to(device).float(),
							  torch.sub(self.margin_mani, 
										torch.masked_select(dist, label_mani[:,0]))
							  )

		loss_nat  = dist_nat.sum()/label_nat_sum if label_nat_sum != 0 else \
					torch.tensor(0).to(device).float()
		loss_mani = dist_mani.sum()/label_mani_sum if label_mani_sum != 0 else \
					torch.tensor(0).to(device).float()
		loss_total = loss_nat + loss_mani

		return loss_total.to(device), loss_mani.to(device), loss_nat.to(device)

	def inference(self, model_output):
		'''output the distance for the final binary mask.'''
		bs, feat_dim, w, h = model_output.size()
		model_output = model_output.permute(0,2,3,1)
		model_output = torch.reshape(model_output, (-1, feat_dim))
		dist = self.pdist(model_output, self.c)
		self.dist = dist
		pred_mask = torch.gt(self.dist, self.threshold).to(torch.float32)
		pred_mask = torch.reshape(pred_mask, (bs,w,h,1)).permute(0,3,1,2)
		self.dist = torch.reshape(self.dist, (bs,w,h,1)).permute(0,3,1,2)
		self.dis_curBatch = pred_mask.to(device).to(torch.float32)
		return self.dis_curBatch.squeeze(dim=1), self.dist.squeeze(dim=1)

def center_radius_init(args, FENet, SegNet, train_data_loader, debug=True, center=None):
	'''the center is the mean-value of pixel features of the real pixels'''
	sample_num = 0
	center = torch.zeros(18).to(device)
	FENet.eval()
	SegNet.eval()
	with torch.no_grad():
		for batch_id, train_data in enumerate(tqdm(train_data_loader, desc="compute center")):  
			image, masks, cls, fcls, scls, tcls = train_data
			if batch_id % 10 != 0:
				continue
			mask_cls = fcls.eq(0)
			image_selected = image[mask_cls,:]
			if image_selected.size()[0] == 0:
				continue
			else:
				sample_num += image_selected.size()[0]
			mask1 = masks[0].to(device)
			image_selected = image_selected.to(device)
			cls = cls.to(device)
			mask1_fea = FENet(image_selected)
			mask1_fea, _, _, _, _, _ = SegNet(mask1_fea, image_selected)
			mask1_fea = torch.mean(mask1_fea,(0,2,3))
			center += mask1_fea

	center = center/sample_num
	pdist  = torch.nn.PairwiseDistance(2)
	radius = torch.tensor(0, dtype=torch.float32).to(device)
	with torch.no_grad():
		for batch_id, train_data in enumerate(tqdm(train_data_loader, desc="compute radius")):  
			if batch_id % 10 != 0:
				continue    
			image, masks, cls, fcls, scls, tcls = train_data
			mask1 = masks[0].to(device)
			image = image.to(device)
			fcls = fcls.to(device)
			mask_cls = fcls.eq(0)
			image_selected = image[mask_cls,:]
			if image_selected.size()[0] == 0:
				continue
			mask1_fea = FENet(image_selected)
			mask1_fea, _, _, _, _, _ = SegNet(mask1_fea, image_selected)
			bs, channel, h, w = mask1_fea.size()
			mask1_fea = mask1_fea.permute(0,2,3,1)      
			mask1_fea = torch.reshape(mask1_fea, (bs*w*h, -1))
			dist = pdist(mask1_fea, center)
			dist_max = torch.max(dist)
			if radius < dist_max:
				radius = dist_max
	return center, radius

def load_center_radius(args, FENet, SegNet, train_data_loader, center_radius_dir='center'):
	'''loading the pre-computed center and radius.'''
	center_radius_path = os.path.join(center_radius_dir, 'radius_center.pth')
	if os.path.exists(center_radius_path):
		load_dict_center_radius = torch.load(center_radius_path)
		center = load_dict_center_radius['center']
		radius = load_dict_center_radius['radius']
		center, radius = center.to(device), radius.to(device)
	else:
		os.makedirs(center_radius_dir, exist_ok=True)
		center, radius = center_radius_init(args, FENet, SegNet, train_data_loader, debug=True)
		torch.save({'center': center, 'radius': radius}, center_radius_path)
	return center, radius

def load_center_radius_api(center_radius_dir='center'):
	'''loading the pre-computed center and radius.'''
	center_radius_path = os.path.join(center_radius_dir, 'radius_center.pth')
	load_dict_center_radius = torch.load(center_radius_path)
	center = load_dict_center_radius['center']
	radius = load_dict_center_radius['radius']
	center, radius = center.to(device), radius.to(device)
	return center, radius


from typing import Dict, Any

import torch
import torch.nn.functional as F
import numpy as np
import torch.nn as nn

from ForensicHub.core.base_model import BaseModel
from ForensicHub.registry import register_model

'''
Synthbuster: Towards detection of diffusion model generated images
'''


def cross_difference(image: torch.Tensor) -> torch.Tensor:
    """
    计算 cross-difference（交叉差分）高通滤波。
    输入: image, shape: [B, C, H, W]
    输出: residual, shape: [B, C, H-1, W-1]
    """
    # Get diagonals
    I = image
    I_tl = I[:, :, :-1, :-1]  # Top-left
    I_br = I[:, :, 1:, 1:]  # Bottom-right
    I_tr = I[:, :, :-1, 1:]  # Top-right
    I_bl = I[:, :, 1:, :-1]  # Bottom-left

    # Cross-difference: abs(I[x,y] + I[x+1,y+1] - I[x,y+1] - I[x+1,y])
    diff = torch.abs(I_tl + I_br - I_tr - I_bl)
    return diff


def fft_features(cross_diff: torch.Tensor, periods=[0, 2, 4, 8]) -> torch.Tensor:
    """
    对交叉差分图像做 FFT 并提取频率点幅值作为特征。
    输入: cross_diff, shape: [B, C, H, W]
    输出: features, shape: [B, C * len(periods) * 2]
    """
    B, C, H, W = cross_diff.shape

    # Compute FFT
    fft = torch.fft.fft2(cross_diff)
    fft_shifted = torch.fft.fftshift(fft, dim=(-2, -1))  # center the zero freq

    # Compute magnitude spectrum
    magnitude = torch.abs(fft_shifted)

    # Center coordinates
    center_h, center_w = H // 2, W // 2

    features = []
    for p in periods:
        for axis in [0, 1]:  # vertical and horizontal
            if axis == 0:
                coords = (center_h + p, center_w)
            else:
                coords = (center_h, center_w + p)
            f = magnitude[:, :, coords[0], coords[1]]  # shape [B, C]
            features.append(f)

    # Concatenate features: [B, C * len(periods) * 2]
    return torch.cat(features, dim=1)


@register_model("Synthbuster")
class Synthbuster(BaseModel):
    def __init__(self):
        super(Synthbuster, self).__init__()
        self.fc1 = nn.Linear(24, 16)
        self.fc2 = nn.Linear(16, 1)

    def forward(self, image, label, **kwargs) -> Dict[str, Any]:
        label = label.float()
        cd = cross_difference(image)
        features = fft_features(cd)
        features = torch.log1p(features)  # log 缩放
        # print(features)
        x = self.fc1(features)
        x = self.fc2(x)
        x = x.squeeze(dim=1) if x.ndim == 2 and x.shape[1] == 1 else x
        combined_loss = F.binary_cross_entropy_with_logits(x, label)
        pred_label = x.sigmoid()

        dict = {
            "backward_loss": combined_loss,
            "pred_label": pred_label,
            "visual_loss": {
                "combined_loss": combined_loss
            }
        }
        return dict



# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------
import os
import time
import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
import matplotlib.pyplot as plt

from torch.optim.lr_scheduler import ReduceLROnPlateau
from kmeans_pytorch import kmeans
from torchvision import transforms
from torch.utils.data import DataLoader
from sklearn import metrics
from torchvision.utils import make_grid
from einops import rearrange
from PIL import Image

Softmax_m = nn.Softmax(dim=1)
device = torch.device('cuda:0')

def device_ids_return(cuda_list):
    '''return the device id'''
    if  len(cuda_list) == 1:
        device_ids = [0]
    elif len(cuda_list) == 2:
        device_ids = [0,1]
    elif len(cuda_list) == 3:
        device_ids = [0,1,2]
    elif len(cuda_list) == 4:
        device_ids = [0,1,2,3]
    elif len(cuda_list) == 7:
        device_ids = [0,1,2,3,4,5,6]
    return device_ids

def findLastCheckpoint(save_dir):
    if os.path.exists(save_dir):
        file_list = os.listdir(save_dir)
        result = 0
        for file in file_list:
            try:
                num = int(file.split('.')[0].split('_')[-1])
                result = max(result, num)
            except:
                continue
        return result
    else:
        os.mkdir(save_dir)
        return 0

def get_confusion_matrix(y_true, y_pred):
    return metrics.confusion_matrix(y_true, y_pred)

def compute_cls_acc_f1(label_lst, pred_lst, iter_num, tb_writer, descr='Val/level3_1'):
    F1  = metrics.f1_score(label_lst, pred_lst, average='macro')
    acc = metrics.accuracy_score(label_lst, pred_lst)
    tb_writer.add_scalar(f'{descr}_F1', F1, iter_num)
    tb_writer.add_scalar(f'{descr}_acc', acc, iter_num)
    print(f"In {descr}, the image-level Acc: {acc:.3f}, F1: {F1:.3f}.")
    print("******************************************************")
    return F1, acc

def tb_writer_display(writer, iter_num, lr_scheduler, epoch, 
                      seg_accu, loc_map_loss, manipul_loss, natural_loss, binary_loss,
                      loss_1, loss_2, loss_3, loss_4):
    writer.add_scalar('Train/seg_accu', seg_accu, iter_num)
    writer.add_scalar('Train/map_loss', loc_map_loss, iter_num)
    writer.add_scalar('Train/binary_map_loss', binary_loss, iter_num)
    writer.add_scalar('Train/manip_loss', manipul_loss, iter_num)
    writer.add_scalar('Train/natur_loss', natural_loss, iter_num)
    writer.add_scalar('Train/loss_1', loss_1, iter_num)
    writer.add_scalar('Train/loss_2', loss_2, iter_num)
    writer.add_scalar('Train/loss_3', loss_3, iter_num)
    writer.add_scalar('Train/loss_4', loss_3, iter_num)
    for count, gp in enumerate(lr_scheduler.optimizer.param_groups,1):
        writer.add_scalar('progress/lr_%d'%count, gp['lr'], iter_num)
    writer.add_scalar('progress/epoch', epoch, iter_num)
    writer.add_scalar('progress/curr_patience',lr_scheduler.num_bad_epochs,iter_num)
    writer.add_scalar('progress/patience',lr_scheduler.patience,iter_num)

def one_hot_label(vector, Softmax_m=Softmax_m):
    x = Softmax_m(vector)
    x = torch.argmax(x, dim=1)
    return x

def one_hot_label_new(vector, Softmax_m=Softmax_m):
    '''
        compute the probability for being as the synthesized image (TODO: double check).
    '''
    x = Softmax_m(vector)
    indices = torch.argmax(x, dim=1)
    prob = 1 - x[:,0]
    indices = list(indices.cpu().numpy())
    prob = list(prob.cpu().numpy())
    return indices, prob

def level_1_convert(input_lst):
    res_lst = []
    for _ in input_lst:
        if _ == 0:
            res_lst.append(0)
        else:
            res_lst.append(1)
    return res_lst

def confusion_matrix_display(label_lst, res_lst, display_lst, display_name):
    confusion_matrix = metrics.confusion_matrix(label_lst, res_lst)
    cm_display = metrics.ConfusionMatrixDisplay(confusion_matrix = confusion_matrix, 
                                                display_labels = display_lst)
    cm_display.plot()
    plt.savefig(f'{display_name}.png')
    confusion_matrix = metrics.confusion_matrix(label_lst, res_lst, normalize='true')
    cm_display = metrics.ConfusionMatrixDisplay(confusion_matrix = confusion_matrix, 
                                                display_labels = display_lst)
    cm_display.plot()
    plt.savefig(f'{display_name}_normalized.png')

def make_folder(folder_name):
    if not os.path.exists(folder_name):
        os.makedirs(folder_name, exist_ok=True)
        print(f"Making folder {folder_name}.")
    else:
        print(f"Folder {folder_name} exists.")

def class_weight(mask, mask_idx):
    '''balance the weight on real and forgery pixel.'''
    mask_balance = torch.ones_like(mask).to(torch.float)
    if (mask == 1).sum():
        mask_balance[mask == 1] = 0.5 / ((mask == 1).sum().to(torch.float) / mask.numel())
        mask_balance[mask == 0] = 0.5 / ((mask == 0).sum().to(torch.float) / mask.numel())
    else:
        pass
        # print(f'Mask{mask_idx} balance is not working!')
    return mask.to(device), mask_balance.to(device)

def setup_optimizer(args, SegNet, FENet):
    '''setup the optimizier, which applies different learning rate on different layers.'''
    '''different hyper-parameters are changed towards HiFi-IFDL dataset.'''
    params_dict_list = []
    params_dict_list.append({'params' : SegNet.module.parameters(), 'lr' : args.learning_rate})
    freq_list = []
    para_list = []
    for name, param in FENet.named_parameters():
        if 'fre' in name:
            freq_list += [param]
        else:
            para_list += [param]
    params_dict_list.append({'params' : freq_list, 'lr' : args.learning_rate*args.lr_backbone})
    params_dict_list.append({'params' : para_list, 'lr' : args.learning_rate})

    optimizer    = torch.optim.Adam(params_dict_list, lr=args.learning_rate*0.75, weight_decay=1e-06)
    lr_scheduler = ReduceLROnPlateau(optimizer, mode='min', factor=args.step_factor, min_lr=1e-08,
                                     patience=args.patience, verbose=True)

    return optimizer, lr_scheduler

def restore_weight_helper(model, model_dir, initial_epoch):
    '''load model given the model_dir that has the model weights.'''
    try:
        weight_path = '{}/{}.pth'.format(model_dir, initial_epoch)
        state_dict = torch.load(weight_path, map_location='cuda:0')['model']
        model.load_state_dict(state_dict)
        print('{} weight-loading succeeds: {}'.format(model_dir, weight_path))
    except:
        print('{} weight-loading fails'.format(model_dir))

    pytorch_total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("{}_params: {}".format(model_dir, pytorch_total_params))
    return model

def restore_optimizer(optimizer, model_dir):
    '''restore the optimizer.'''
    try:
        weight_path = '{}/{}.pth'.format(model_dir, initial_epoch)
        state_dict = torch.load(weight_path, map_location='cuda:0')
        print('Optimizer weight-loading succeeds.')
        optimizer.load_state_dict(state_dict['optimizer'])
    except:
        # print('{} Optimizer weight-loading fails.')
        pass
    return optimizer

def composite_obj(args, loss, loss_1, loss_2, loss_3, loss_4, loss_binary):
    ''' 'base', 'fg', 'local', 'full' '''
    if args.ablation == 'full':     # fine-grained + localization
        loss_total = 100*loss + loss_1 + loss_2 + loss_3 + 100*loss_4 + loss_binary
    elif args.ablation == 'base':   # one-shot
        loss_total = loss_4
    elif args.ablation == 'fg':     # only fine-grained
        loss_total = loss_1 + loss_2 + loss_3 + loss_4
    elif args.ablation == 'local':  # only loclization
        loss_total = loss + 10e-6*(loss_1 + loss_2 + loss_3 + loss_4)
    else:
        assert False
    return loss_total

def composite_obj_step(args, loss_4_sum, map_loss_sum):
    ''' return loss for the scheduler '''
    if args.ablation == 'full':
        schedule_step_loss = loss_4_sum + map_loss_sum
    elif args.ablation == 'base':
        schedule_step_loss = loss_4_sum
    elif args.ablation == 'fg':
        schedule_step_loss = loss_4_sum
    elif args.ablation == 'local':
        schedule_step_loss = map_loss_sum
    else:
        assert False
    return schedule_step_loss

def viz_log(args, mask, pred_mask, image, iter_num, step, mode='train'):
    '''viz training, val and inference.'''
    mask = torch.unsqueeze(mask, dim=1)
    pred_mask = torch.unsqueeze(pred_mask, dim=1)
    mask_viz = torch.cat([mask]*3, axis=1)
    pred_mask = torch.cat([pred_mask]*3, axis=1)
    image = torch.nn.functional.interpolate(image,  # for viz.
                                          size=(256, 256), 
                                          mode='bilinear')
    fig_viz = torch.cat([mask_viz, image, pred_mask], axis=0)
    grid = make_grid(fig_viz, nrow=mask_viz.shape[0])   # nrow in fact is the column number.
    grid = 255. * rearrange(grid, 'c h w -> h w c').cpu().numpy()
    img_h = Image.fromarray(grid.astype(np.uint8))
    # os.makedirs(f"./viz_{mode}_{args.learning_rate}/", exist_ok=True)
    os.makedirs(f"./viz_{mode}/", exist_ok=True)
    if mode == 'train':
        # img_h.save(f"./viz_{mode}_{args.learning_rate}/iter_{iter_num}.jpg")
        img_h.save(f"./viz_{mode}/iter_{iter_num}.jpg")
    else:
        # img_h.save(f"./viz_{mode}_{args.learning_rate}/iter_{iter_num}_step_{step}.jpg")
        img_h.save(f"./viz_{mode}/iter_{iter_num}_step_{step}.jpg")

def process_mask(mask, pred_mask):
    '''process the mask'''
    pred_mask = torch.unsqueeze(pred_mask, dim=1)
    mask = torch.unsqueeze(mask, dim=1)
    pred_mask = torch.cat([pred_mask]*3, axis=1)
    mask = torch.cat([mask]*3, axis=1)

    pred_mask = nn.functional.interpolate(pred_mask, 
                                        size=(256, 256), mode='bilinear')
    mask = nn.functional.interpolate(mask, 
                                    size=(256, 256), mode='bilinear')

    return pred_mask, mask

def viz_logs_scale(args, iter_num, mask_128, mask_64, mask_32, mask2, mask3, mask4, mode='train'):
    '''visualize the mask and predicted mask.'''
    pred_mask_128, mask128 = process_mask(mask_128, mask2)
    pred_mask_64, mask64 = process_mask(mask_64, mask3)
    pred_mask_32, mask32 = process_mask(mask_32, mask4)

    fig_viz = torch.cat([pred_mask_32, mask32, pred_mask_64, mask64, 
                        pred_mask_128, mask128], axis=0)
    grid = make_grid(fig_viz, nrow=pred_mask_32.shape[0])
    grid = 255. * rearrange(grid, 'c h w -> h w c').cpu().numpy()
    img_h = Image.fromarray(grid.astype(np.uint8))
    os.makedirs(f"./viz_{mode}_{args.learning_rate}/", exist_ok=True)
    img_h.save(f"./viz_{mode}_{args.learning_rate}/iter_{iter_num}_pred.jpg")

def train_log_dump(args, seg_correct, seg_total, map_loss_sum, mani_lss_sum, natu_lss_sum,
                    binary_map_loss_sum, loss_1_sum, loss_2_sum, loss_3_sum,
                    loss_4_sum, epoch, step, writer, iter_num, lr_scheduler):
    '''compute and output the different training loss & seg accuarcy.'''
    seg_accu = seg_correct / seg_total * 100
    loc_map_loss = map_loss_sum / args.dis_step
    manipul_loss = mani_lss_sum / args.dis_step
    natural_loss = natu_lss_sum / args.dis_step
    binary_loss  = binary_map_loss_sum / args.dis_step
    loss_1 = loss_1_sum / args.dis_step
    loss_2 = loss_2_sum / args.dis_step
    loss_3 = loss_3_sum / args.dis_step
    loss_4 = loss_4_sum / args.dis_step
    print(f'[Epoch: {epoch+1}, Step: {step + 1}] batch_loc_acc: {seg_accu:.2f}')
    print(f'cls1_loss: {loss_1:.3f}, cls2_loss: {loss_2:.3f}, cls3_loss: {loss_3:.3f}, '+
          f'cls4_loss: {loss_4:.3f}, map_loss:   {loc_map_loss:.3f}, '+
          f'manip_loss: {manipul_loss:.3f}, natur_loss: {natural_loss:.3f}, '+
          f'binary_map_loss: {binary_loss:.3f}') 
    '''write the tensorboard.'''
    tb_writer_display(writer, iter_num, lr_scheduler, epoch, seg_accu, 
                      loc_map_loss, manipul_loss, natural_loss, binary_loss,
                      loss_1, loss_2, loss_3, loss_4)


import sys
import torch
from PIL import Image
from torchvision import transforms
from ForensicHub.registry import DATASETS, MODELS, POSTFUNCS, TRANSFORMS, build_from_registry


def load_image(image_path, image_size, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
    transform = transforms.Compose([
        transforms.Resize((image_size, image_size)),
        transforms.ToTensor(),
        transforms.Normalize(mean=mean, std=std)
    ])
    image = Image.open(image_path).convert("RGB")
    image = transform(image).unsqueeze(0)  # (1, 3, H, W)
    return image


if __name__ == '__main__':
    # Set model configs here
    model_args = {
        "name": "ConvNextSmall",
        # model init config
        "init_config": {
            "image_size": 256
        },
        # model pretrained weight path
        "init_path": "/mnt/data1/dubo/workspace/ForensicHub/log/crossdataset_image_new/crossdataset_convnextsmall_none_train/checkpoint-19.pth"
    }

    model = build_from_registry(MODELS, model_args)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    # Load weights
    checkpoint = torch.load(model_args['init_path'], map_location=device)
    model.load_state_dict(checkpoint["model"] if "model" in checkpoint else checkpoint)

    model.eval()

    # Load image
    image_path = '/mnt/data3/public_datasets/AIGC/GenImage/ADM/imagenet_ai_0508_adm/train/ai/375_adm_64.PNG'
    image_size = 256
    image_tensor = load_image(image_path=image_path, image_size=image_size).to(device)

    input_dict = {
        'image': image_tensor,
        # pseudo input
        "label": torch.ones(1).to(device),
        "mask": torch.ones(1, 1, image_size, image_size).to(device),
    }
    out_dict = model(**input_dict)
    pred_label = out_dict['pred_label'].item()

    print(f"Predicted label: {pred_label}")



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union
import os
import torch
from torch.utils.data import Dataset


class BaseDataset(Dataset, ABC):
    """Base class for all forensic datasets in ForensicHub.
    
    This class defines the interface that all dataset classes must implement.
    It inherits from torch.utils.data.Dataset and adds forensic-specific functionality.
    """

    def __init__(self,
                 path: Union[str, List[str]],
                 common_transform: Optional[Any] = None,
                 post_transform: Optional[Any] = None,
                 img_loader: Any = None,
                 post_funcs: Any = None,
                 **kwargs):
        """Initialize the dataset.
        
        Args:
            path (str): Path to the dataset.
            transform (Optional[Any]): Transform to be applied to the input data.
            img_loader (Any): Function to load images.
            post_funcs (Optional[List[callable]]): List of post-processing functions.
            **kwargs: Additional arguments specific to each dataset.
        """
        super().__init__()
        self.path = path
        self.common_transform = common_transform
        self.post_transform = post_transform
        self.img_loader = img_loader
        self.post_funcs = post_funcs

        # Initialize dataset paths
        self._init_dataset_path()

    @abstractmethod
    def _init_dataset_path(self) -> None:
        """Initialize dataset paths.
        
        This method should be implemented by each dataset class to set up
        the paths for images and labels/masks.
        """
        pass

    def __len__(self) -> int:
        """Return the number of samples in the dataset."""
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        """Get a sample from the dataset.
        
        Args:
            idx (int): Index of the sample to get.
            
        Returns:
            Dict[str, torch.Tensor]: Dictionary containing sample data.
            The dictionary should at least contain:
            - 'image': torch.Tensor, the input image
            Optional fields:
            - 'mask': torch.Tensor, the target mask
            - 'label': torch.Tensor, the target label
            - 'origin_shape': torch.Tensor, the original image shape
            - 'edge_mask': torch.Tensor, the edge mask of the image
            - ......
        """
        pass

    def __str__(self) -> str:
        """Return string representation of the dataset."""
        cls_name = self.__class__.__name__
        cls_path = self.path
        cls_len = len(self)
        return f"[{cls_name}] at {cls_path}, with length of {cls_len:,}"



# ------------------------------------------------------------------------------
# Author: Xiao Guo (guoxia11@msu.edu)
# CVPR2023: Hierarchical Fine-Grained Image Forgery Detection and Localization
# ------------------------------------------------------------------------------

import torch
import torch.nn as nn
import torch.nn.functional as F
from .seg_hrnet_config import get_cfg_defaults
import time


def weights_init(init_type='gaussian'):
    def init_fun(m):
        classname = m.__class__.__name__
        if (classname.find('Conv') == 0 or classname.find(
                'Linear') == 0) and hasattr(m, 'weight'):
            if init_type == 'gaussian':
                nn.init.normal_(m.weight, 0.0, 0.02)
            elif init_type == 'xavier':
                nn.init.xavier_normal_(m.weight, gain=math.sqrt(2))
            elif init_type == 'kaiming':
                nn.init.kaiming_normal_(m.weight, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                nn.init.orthogonal_(m.weight, gain=math.sqrt(2))
            elif init_type == 'default':
                pass
            else:
                assert 0, "Unsupported initialization: {}".format(init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                nn.init.constant_(m.bias, 0.0)

    return init_fun


class PartialConv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super().__init__()
        self.input_conv = nn.Conv2d(in_channels, out_channels, kernel_size,
                                    stride, padding, dilation, groups, bias)
        self.mask_conv = nn.Conv2d(in_channels, out_channels, kernel_size,
                                   stride, padding, dilation, groups, False)
        self.input_conv.apply(weights_init('kaiming'))
        torch.nn.init.constant_(self.mask_conv.weight, 1.0)
        # mask is not updated
        for param in self.mask_conv.parameters():
            param.requires_grad = False

    def forward(self, input, mask):
        # http://masc.cs.gmu.edu/wiki/partialconv
        # C(X) = W^T * X + b, C(0) = b, D(M) = 1 * M + 0 = sum(M)
        # W^T* (M .* X) / sum(M) + b = [C(M .* X) – C(0)] / D(M) + C(0)

        ## GX: masking the input outside function.
        output = self.input_conv(input)
        if self.input_conv.bias is not None:
            output_bias = self.input_conv.bias.view(1, -1, 1, 1).expand_as(output)
        else:
            output_bias = torch.zeros_like(output)

        with torch.no_grad():
            output_mask = self.mask_conv(mask)

        no_update_holes = output_mask == 0

        ## in output_mask, fills the 0-value-position with 1.0
        ## without this step, math error occurs.
        mask_sum = output_mask.masked_fill_(no_update_holes, 1.0)
        output_pre = (output - output_bias) / mask_sum + output_bias
        output = output_pre.masked_fill_(no_update_holes, 0.0)

        new_mask = torch.ones_like(output)
        new_mask = new_mask.masked_fill_(no_update_holes, 0.0)

        return output, new_mask


class NonLocalMask(nn.Module):
    def __init__(self, in_channels, reduce_scale):
        super(NonLocalMask, self).__init__()

        self.r = reduce_scale

        # input channel number
        self.ic = in_channels * self.r * self.r

        # middle channel number
        self.mc = self.ic

        self.g = nn.Conv2d(in_channels=self.ic, out_channels=self.ic,
                           kernel_size=1, stride=1, padding=0)

        self.theta = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                               kernel_size=1, stride=1, padding=0)
        self.phi = nn.Conv2d(in_channels=self.ic, out_channels=self.mc,
                             kernel_size=1, stride=1, padding=0)
        self.W_s = nn.Conv2d(in_channels=in_channels, out_channels=in_channels,
                             kernel_size=1, stride=1, padding=0)

        self.gamma_s = nn.Parameter(torch.ones(1))
        self.getmask = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=16,
                      kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(in_channels=16, out_channels=1, kernel_size=3, stride=1, padding=1)
        )

        ## Pconv
        self.Pconv_1 = PartialConv(3, 3, kernel_size=3, stride=2)
        self.Pconv_2 = PartialConv(3, 3, kernel_size=3, stride=2)
        self.Pconv_3 = PartialConv(3, 1, kernel_size=3, stride=2)

    def forward(self, x, img):
        b, c, h, w = x.shape

        x1 = x.reshape(b, self.ic, h // self.r, w // self.r)

        # g x
        g_x = self.g(x1).view(b, self.ic, -1)
        g_x = g_x.permute(0, 2, 1)

        # theta
        theta_x = self.theta(x1).view(b, self.mc, -1)
        theta_x_s = theta_x.permute(0, 2, 1)

        # phi x
        phi_x = self.phi(x1).view(b, self.mc, -1)
        phi_x_s = phi_x

        # non-local attention
        f_s = torch.matmul(theta_x_s, phi_x_s)
        f_s_div = F.softmax(f_s, dim=-1)

        # get y_s
        y_s = torch.matmul(f_s_div, g_x)
        y_s = y_s.permute(0, 2, 1).contiguous()
        y_s = y_s.view(b, c, h, w)

        # GX: (256,256,18), output mask for the deep metric loss.
        mask_feat = x + self.gamma_s * self.W_s(y_s)

        # get 1-dimensional mask_tmp
        mask_binary = torch.sigmoid(self.getmask(mask_feat))
        mask_tmp = mask_binary.repeat(1, 3, 1, 1)
        mask_img = img * mask_tmp  # mask_img is the overlaid image.

        ## conv output
        x, new_mask = self.Pconv_1(mask_img, mask_tmp)
        x, new_mask = self.Pconv_2(x, new_mask)
        x, _ = self.Pconv_3(x, new_mask)
        mask_binary = mask_binary.squeeze(dim=1)
        return x, mask_feat, mask_binary


class Flatten(nn.Module):
    def __init__(self):
        super(Flatten, self).__init__()

    def forward(self, x):
        return x.view(x.size(0), -1)


class Classifer(nn.Module):
    def __init__(self, in_channels, output_channels):
        super(Classifer, self).__init__()
        self.pool = nn.Sequential(
            # nn.AdaptiveAvgPool2d((1,1)),
            nn.AdaptiveAvgPool2d(1),
            Flatten()
        )
        self.fc = nn.Linear(in_channels, output_channels, bias=True)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        feat = self.pool(x)
        feat = self.relu(feat)
        cls_res = self.fc(feat)
        return cls_res


class BranchCLS(nn.Module):
    def __init__(self, in_channels, output_channels):
        super(BranchCLS, self).__init__()
        self.pool = nn.Sequential(nn.AdaptiveAvgPool2d((1, 1)),
                                  Flatten()
                                  )
        self.fc = nn.Linear(18, output_channels, bias=True)
        self.bn = nn.BatchNorm1d(18, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        self.branch_cls = nn.Sequential(nn.Conv2d(in_channels=in_channels, out_channels=32,
                                                  padding=1, kernel_size=3, stride=1),
                                        nn.ReLU(inplace=True),
                                        nn.Conv2d(in_channels=32, out_channels=18,
                                                  padding=1, kernel_size=3, stride=1),
                                        nn.ReLU(inplace=True),
                                        )
        self.leakyrelu = nn.LeakyReLU(0.2)

    def forward(self, x):
        feat = self.branch_cls(x)
        x = self.pool(feat)
        x = self.bn(x)
        cls_res = self.fc(x)
        cls_pro = self.leakyrelu(cls_res)
        zero_vec = -9e15 * torch.ones_like(cls_pro)
        cls_pro = torch.where(cls_pro > 0, cls_pro, zero_vec)
        return cls_res, cls_pro, feat


class NLCDetection(nn.Module):
    def __init__(self):
        super(NLCDetection, self).__init__()
        self.split_tensor_1 = torch.tensor([1, 3]).cuda()
        self.split_tensor_2 = torch.tensor([1, 2, 1, 3]).cuda()
        self.softmax_m = nn.Softmax(dim=1)
        FENet_cfg = get_cfg_defaults()
        feat1_num, feat2_num, feat3_num, feat4_num = FENet_cfg['STAGE4']['NUM_CHANNELS']

        ## mask generation branch.
        self.getmask = NonLocalMask(feat1_num, 4)

        ## classification branch.
        self.branch_cls_level_1 = BranchCLS(271, 14)  # 252 + 18 = 270
        self.branch_cls_level_2 = BranchCLS(252, 7)  # 144+72+36 = 252
        self.branch_cls_level_3 = BranchCLS(216, 5)  # 144+72 = 216
        self.branch_cls_level_4 = BranchCLS(144, 3)  # 144

    def forward(self, feat, img):
        s1, s2, s3, s4 = feat

        pconv_feat, mask, mask_binary = self.getmask(s1, img)
        pconv_feat = pconv_feat.clone().detach()

        pconv_1 = F.interpolate(pconv_feat, size=s1.size()[2:], mode='bilinear', align_corners=True)

        ## forth branch.
        cls_4, pro_4, _ = self.branch_cls_level_4(s4)
        cls_prob_4 = self.softmax_m(pro_4)
        cls_prob_40 = torch.unsqueeze(cls_prob_4[:, 0], 1)
        cls_prob_41 = torch.unsqueeze(cls_prob_4[:, 1], 1)
        cls_prob_42 = torch.unsqueeze(cls_prob_4[:, 2], 1)
        cls_prob_mask_3 = torch.cat([cls_prob_40, cls_prob_41, cls_prob_41, cls_prob_42, cls_prob_42], axis=1)

        ## third branch
        s4F = F.interpolate(s4, size=s3.size()[2:], mode='bilinear', align_corners=True)
        s3_input = torch.cat([s4F, s3], axis=1)
        cls_3, pro_3, _ = self.branch_cls_level_3(s3_input)
        cls_prob_3 = self.softmax_m(pro_3)
        cls_3 = cls_3 + cls_3 * cls_prob_mask_3
        cls_prob_30 = torch.unsqueeze(cls_prob_3[:, 0], 1)
        cls_prob_31 = torch.unsqueeze(cls_prob_3[:, 1], 1)
        cls_prob_32 = torch.unsqueeze(cls_prob_3[:, 2], 1)
        cls_prob_33 = torch.unsqueeze(cls_prob_3[:, 3], 1)
        cls_prob_34 = torch.unsqueeze(cls_prob_3[:, 4], 1)
        cls_prob_mask_2 = torch.cat([cls_prob_30, cls_prob_31, cls_prob_31,
                                     cls_prob_32, cls_prob_32,
                                     cls_prob_33, cls_prob_34], axis=1)

        ## second branch
        s3F = F.interpolate(s3_input, size=s2.size()[2:], mode='bilinear', align_corners=True)
        s2_input = torch.cat([s3F, s2], axis=1)
        cls_2, pro_2, _ = self.branch_cls_level_2(s2_input)
        cls_prob_2 = self.softmax_m(pro_2)
        cls_2 = cls_2 + cls_2 * cls_prob_mask_2
        cls_prob_20 = torch.unsqueeze(cls_prob_2[:, 0], 1)
        cls_prob_21 = torch.unsqueeze(cls_prob_2[:, 1], 1)
        cls_prob_22 = torch.unsqueeze(cls_prob_2[:, 2], 1)
        cls_prob_23 = torch.unsqueeze(cls_prob_2[:, 3], 1)
        cls_prob_24 = torch.unsqueeze(cls_prob_2[:, 4], 1)
        cls_prob_25 = torch.unsqueeze(cls_prob_2[:, 4], 1)
        cls_prob_26 = torch.unsqueeze(cls_prob_2[:, 4], 1)
        cls_prob_mask_1 = torch.cat([cls_prob_20,
                                     cls_prob_21, cls_prob_21, cls_prob_22, cls_prob_22,  # 4 diffusion
                                     cls_prob_23, cls_prob_23, cls_prob_24, cls_prob_24,  # 4 gan
                                     cls_prob_25, cls_prob_25,  # faceshifter+stgan
                                     cls_prob_26, cls_prob_26, cls_prob_26], axis=1)  # 3 editing

        s2F = F.interpolate(s2_input, size=s1.size()[2:], mode='bilinear', align_corners=True)
        s1_input = torch.cat([s2F, s1, pconv_1], axis=1)
        cls_1, pro_1, _ = self.branch_cls_level_1(s1_input)
        cls_1 = cls_1 + cls_1 * cls_prob_mask_1
        return mask, mask_binary, cls_4, cls_3, cls_2, cls_1
